#!/usr/bin/env python2
# coding=utf-8

# Copyright (c) 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function
import random
import abc
import collections
import struct
import sys
from textwrap import dedent
from mako.template import Template

# Module-local global variable that tracks all of the structure types that
# have been generated since ever.
struct_types = dict()

ALL130_TYPES = [
    "float",  "vec2",  "vec3",  "vec4",
    "int",    "ivec2", "ivec3", "ivec4",
    "uint",   "uvec2", "uvec3", "uvec4",
    "bool",   "bvec2", "bvec3", "bvec4",

    "mat2",    "mat2x3",  "mat2x4",
    "mat3x2",  "mat3",    "mat3x4",
    "mat4x2",  "mat4x3",  "mat4"
]

DOUBLE_TYPES = [
    "double", "dvec2", "dvec3", "dvec4",

    "dmat2",   "dmat2x3", "dmat2x4",
    "dmat3x2", "dmat3",   "dmat3x4",
    "dmat4x2", "dmat4x3", "dmat4"
]

ALL400_TYPES = ALL130_TYPES + DOUBLE_TYPES

# All known types, including the redundant NxN matrix types.
ALL_TYPES = [ "mat2x2",  "mat3x3",  "mat4x4",
             "dmat2x2", "dmat3x3", "dmat4x4"] + ALL400_TYPES

def align(offset, alignment):
    return ((offset + alignment - 1) / alignment) * alignment


def isscalar(type):
    """Return true if the type is a known scalar type from any GLSL version."""
    return type in ["float", "bool", "int", "uint", "double"]


def isvector(type):
    """Return true if the type is a known vector type from any GLSL version."""
    return type in ["vec2",  "vec3",  "vec4",
                    "ivec2", "ivec3", "ivec4",
                    "uvec2", "uvec3", "uvec4",
                    "bvec2", "bvec3", "bvec4",
                    "dvec2", "dvec3", "dvec4"]


def ismatrix(type):
    """Return true if the type is a known matrix type from any GLSL version."""
    return type in ["mat2",    "mat3",    "mat4",
                    "mat2x2",  "mat2x3",  "mat2x4",
                    "mat3x2",  "mat3x3",  "mat3x4",
                    "mat4x2",  "mat4x3",  "mat4x4",
                    "dmat2",   "dmat3",   "dmat4",
                    "dmat2x2", "dmat2x3", "dmat2x4",
                    "dmat3x2", "dmat3x3", "dmat3x4",
                    "dmat4x2", "dmat4x3", "dmat4x4"]


def isarray(type):
    """Return true if the type name appears to be an array type."""
    return "[" in type


def isstructure(type):
    """Return true if the type name appears to be a structure type.

    This function works by process of elimination.  If the named type is not a
    scalar, vector, matrix, or array, then it must be a structure.

    NOTE: If the type is an array-of-structure type, this function will return
    false.
    """
    return not (isscalar(type) or isvector(type) or ismatrix(type) or
                isarray(type))


def vector_size(type):
    """Return the number of elements in a vector type."""
    if isvector(type):
        return int(type[-1:])

    raise Exception("Non-vector type {}".format(type))


def matrix_dimensions(type):
    """Return the dimensions of an array type.

    The tuple returned is (columns, rows).
    """
    if not ismatrix(type):
        raise Exception("Non-matrix type {}".format(type))

    if "x" in type:
        s = type[-3:].split("x")
        return int(s[0]), int(s[1])
    else:
        d = int(type[-1:])
        return d, d


def array_elements(type):
    """Return the number of elements in an array type.

    For non-array types zero is returned.
    """
    if "[" not in type:
        return 0

    # Is there a better way to do this?
    return int(type.split("[")[1].split("]")[0])

def basic_machine_units(type):
    """Return the size in 'basic machine units' of a scalar type."""
    if type in ["float", "bool", "int", "uint"]:
        return 4

    if type == "double":
        return 8

    raise Exception("Non-scalar type {}".format(type))


def array_base_type(type):
    """Return the type of an element from an array type."""
    if not isarray(type):
        raise Exception("Non-array type {}".format(type))

    return type[:type.index("[")] + type[type.index("]")+1:]

def without_array(type):
    """Return the type of an element with all arrays removed."""
    if not isarray(type):
        raise Exception("Non-array type {}".format(type))

    return type.split("[")[0]

def arrays_of_arrays_size(type):
    """Return the total number of elements in an array of arrays."""
    if not isarray(type):
        raise Exception("Non-array type {}".format(type))

    size = 1
    base_type = type
    while (isarray(base_type)):
        size = size * array_elements(base_type)
        base_type = array_base_type(base_type)

    return size

def vector_base_type(type):
    """Return the type of an element from a vector."""
    if not isvector(type):
        raise Exception("Non-vector type {}".format(type))

    if type[0] == 'v':
        return "float"
    elif type[0] == 'i':
        return "int"
    elif type[0] == 'u':
        return "uint"
    elif type[0] == 'b':
        return "bool"
    elif type[0] == 'd':
        return "double"
    else:
        raise Exception("Unknown vector type {}".format(type))


def iterate_structures(fields, types_seen=[], types_yielded=[]):
    """Given a list of fields, yields the structures in the fields in proper
    declaration order.  Detects recurrsion in the types and raises an
    exception.
    """
    for (type, name) in fields:
        if isarray(type):
            type = without_array(type)

        if not isstructure(type):
            continue

        if type in types_seen:
            raise Exception("Type recurrsion involving {}".format(type))

        for t in iterate_structures(struct_types[type],
                                    types_seen + [type],
                                    types_yielded):
            yield t

        if type not in types_yielded:
            types_yielded.append(type)
            yield type


def select_basic_type(types, names):
    """Return a random type with a matching field name.

    From the list of supplied types, picks a random type.  Using supplied
    unique_name_dict, a name is also chosen.  The two are returned as a tuple.
    """
    t = random.choice(types)
    return t, names.get_name(t)


def generate_struct_of_basic_types(types, names):
    """Return a sequence of random types with unique field names."""
    return [select_basic_type(types, names)
            for i in xrange(0, random.randint(1,12))]


def generate_member_from_description(description, builtin_types, names):
    """Generate a single block member for a list of required attributes.

    The list of required attributes is a sequence of 'struct' and 'array'
    perhaps terminated with the name of a GLSL basic type.  In addition to any
    required subfields, structures generated in the sequence will be populated
    with a random selection of available GLSL types named in builtin_types.

    A matrix orientation can also be listed as the first requirement.
    However, that is not handled by this function, and it will be ignored.

    The tuple returned is (member_type, member_name).
    """
    if len(description) == 0:
        return select_basic_type(builtin_types, names)

    item = description[0]
    if item == "array":
        base_type, name = generate_member_from_description(
            description[1:],
            builtin_types,
            names)

        # If we're making an array of something that can be "big," try to make
        # the array a little smaller.

        if ismatrix(base_type) or isarray(base_type) or isstructure(base_type):
            size = random.choice([2, 3, 5, 7])
        else:
            size = random.choice([3, 5, 7, 11, 13])

        t = "{}[{}]".format(base_type, size)
        return t, name
    elif item == "struct":
        fields = generate_struct_of_basic_types(builtin_types, names)
        random.shuffle(fields)

        # Peek ahead.  If the next item in the description is a built-in type,
        # then all of the remaining items must be built-in types.  Generate a
        # list of these.

        if len(description) > 1 and description[1] in ALL_TYPES:
            required_fields = [generate_member_from_description([i],
                                                                builtin_types,
                                                                names)
                               for i in description[1:]]

        else:
            required_fields = [generate_member_from_description(
                    description[1:],
                    builtin_types,
                    names)]

        # Pick a random spot in the list of "common" fields and insert all of
        # the required fields there.

        j = random.randint(0, len(fields))
        f = fields[:j] + required_fields + fields[j:]

        struct_name = "S{}".format(len(struct_types) + 1)
        struct_types[struct_name] = f

        field_name = names.get_name(struct_name)
        return struct_name, field_name
    elif item in ALL_TYPES:
        return item, names.get_name(item)
    elif item in ["row_major", "column_major", "#column_major"]:
        # While "row_major" and "column_major" are valid requirements, they
        # are not processed here.  Just skip over them for now.
        return generate_member_from_description(description[1:],
                                                builtin_types,
                                                names)

    raise Exception("Invalid UBO member description {}".format(item))


def generate_ubo(description_list, builtin_types):
    """Generate a uniform block for a list of requirements.

    Each member of description_list is taken to be a sequence of requirements
    for a single field in the uniform block.  These are processed by
    generate_member_from_description.

    In addition to the list of required fields, the uniform block is also
    populated with a random selection of available GLSL types named in
    builtin_types.
    """
    layouts = dict()
    names = unique_name_dict()

    fields = []

    for desc in description_list:
        m = generate_member_from_description(desc, builtin_types, names)
        fields.append(m)

        if desc[0] in ["row_major", "column_major", "#column_major"]:
            layouts[m[1]] = desc[0]

    fields.extend(generate_struct_of_basic_types(builtin_types, names))
    random.shuffle(fields)

    required_layouts = []
    for (field_type, field_name) in fields:
        if field_name in layouts:
            required_layouts.append(layouts[field_name])
        else:
            required_layouts.append(None)

    return fields, required_layouts


def generate_layouts(fields, required_layouts, allow_row_major_structure):
    """Generate a directed random sequence of layout qualifiers.

    Elements in requred_layouts that are non-None will have the same value in
    the final sequence.  Remaining elements matching entries in fields that
    are either structure or matrix types will have random layouts assigned.

    If allow_row_major_structure is not set, only matrix types will have
    random layout assignments.  Explicit layouts in required_layouts are
    unaffected.
    """
    if required_layouts is None:
        required_layouts = [None] * len(fields)

    layouts = []
    for ((type, name), lay) in zip(fields, required_layouts):
        if isarray(type):
            type = without_array(type)

        if lay:
            layouts.append(lay)
        elif isstructure(type) and not allow_row_major_structure:
            # This would work-around a bug in NVIDIA closed source drivers.
            # They do not propagate row-major down into structures.

            layouts.append("#column_major")
        elif ismatrix(type) or isstructure(type):
            # Choose a random matrix orientation.  The #column_major are
            # ignored when the UBO is emitted, but when a the UBO is
            # re-emitted with a default row-major layout, these become
            # "column_major".

            layouts.append(random.choice(["#column_major",
                                          "#column_major",
                                          "#column_major",
                                          "row_major",
                                          "row_major",
                                          "column_major"]))
        else:
            layouts.append("")
    return layouts


def relayout_for_default_row_major(l):
    """Return a new layout for an assumed default layout of row-major."""
    if l == "row_major":
        return "#row_major"
    elif l == "column_major" or l == "#column_major":
        return "column_major"
    elif l == "":
        return ""
    else:
        raise Exception("Invalid layout {}".format(l))


def fields_to_glsl_struct(type):
    """From a name structure type, generate a GLSL definition of that
    structure.
    """
    # The longest type name will have the form 'dmatCxR[##]' for 11
    # characters.  Use this to set the spacing between the field type and the
    # field name.

    structure_template = Template(dedent("""\
    struct ${struct_name} {
    % for (field_type, field_name) in fields:
        ${"{:<11}".format(field_type)} ${field_name};
    % endfor
    };
    """))

    return structure_template.render(struct_name=type, fields=struct_types[type])

def iterate_struct_array_recursive(field_type,
                                   name_from_API,
                                   name_from_shader,
                                   packing,
                                   offset,
                                   row_major,
                                   field_row_major,
                                   explicit_layout):

    astride = packing.array_stride(field_type, field_row_major)
    array_member_align = packing.base_alignment(field_type, field_row_major)

    element_t = array_base_type(field_type)

    for i in xrange(array_elements(field_type)):
        name_from_API_with_index = "{}[{}]".format(name_from_API, i)
        name_from_shader_with_index = "{}[{}]".format(name_from_shader, i)

        if isarray(element_t):
            inner_aoa_size = arrays_of_arrays_size(element_t)
            o = align(offset, array_member_align) + (astride * i * inner_aoa_size)

            for x in iterate_struct_array_recursive(element_t,
                                                    name_from_API_with_index,
                                                    name_from_shader_with_index,
                                                    packing,
                                                    o,
                                                    row_major,
                                                    field_row_major,
                                                    explicit_layout):
                yield x

                a = packing.base_alignment(x.GLSL_type, row_major)
                o = align(o, a) + packing.size(x.GLSL_type, row_major)
        else:
            o = align(offset, array_member_align) + (astride * i)
            yield block_member(
                name_from_shader_with_index,
                name_from_API_with_index,
                element_t,
                explicit_layout,
                o,
                field_row_major)
            for x in iterate_all_recursive(struct_types[element_t],
                                           None,
                                           name_from_API_with_index,
                                           name_from_shader_with_index,
                                           packing,
                                           o,
                                           field_row_major):
                yield x

                a = packing.base_alignment(x.GLSL_type, row_major)
                o = align(o, a) + packing.size(x.GLSL_type, row_major)

def iterate_all_recursive(fields,
                          field_layouts,
                          name_from_API_base,
                          name_from_shader_base,
                          packing,
                          offset,
                          row_major):
    """In-order recursive iteration of all subfields of a structure type

    fields -- List of fields in the uniform block or structure
    field_layouts -- Layout for each field in the uniform block (None for
        structures).
    name_from_API_base -- Base name of the member as it would be accessed via
        the GL API.
    name_from_shader_base -- Base name of the member as it would be accessed via
        the GLSL shader.
    packing -- Packing rules used to layout the type
    offset -- Aligned base offset of the member
    row_major -- True if the current matrix orientation is row-major

    At each step of the iteration will yield a block_member object
    representing the member.

    Basic GLSL types will be visited once.  Arrays of basic GLSL types will
    also be visited once.

    A structure will be visited once, then each field in the structure will be
    visited.

    An array of structures will be visited once, then each element of the
    array will be visited.  Since each element is a structure, each field will
    be visited.

    Consider the following structure:

        struct S2 {
            int i;
            float f[2];
        };

        struct S1 {
            S2 s[2];
        };

    If name_from_API_base is "x", the members would be visited in the
    following order: "x.s", "x.s[0]", "x.s[0].i", "x.s[0].f[0]", "x.s[1]",
    "x.s[1].i", "x.s[1].f[0]".

    See also iterate_all_block_members.
    """
    if field_layouts is None:
        field_layouts = [""] * len(fields)

    if len(name_from_shader_base) > 0:
        fmt = "{base}.{field}"
    else:
        fmt = "{field}"

    for ((field_type, field_name), explicit_layout) in zip(fields,
                                                           field_layouts):
        name_from_shader = fmt.format(
            base=name_from_shader_base,
            field=field_name)
        name_from_API = fmt.format(
            base=name_from_API_base,
            field=field_name)

        if explicit_layout == "row_major":
            field_row_major = True
        elif explicit_layout == "column_major":
            field_row_major = False
        else:
            field_row_major = row_major

        if isarray(field_type):
            base_type = without_array(field_type)

            if isstructure(base_type):
                yield block_member(
                    name_from_shader,
                    name_from_API,
                    field_type,
                    explicit_layout,
                    offset,
                    field_row_major)

                for x in iterate_struct_array_recursive(field_type,
                                           name_from_API,
                                           name_from_shader,
                                           packing,
                                           offset,
                                           row_major,
                                           field_row_major,
                                           explicit_layout):
                    yield x

            elif ismatrix(base_type):
                yield block_member(
                    name_from_shader,
                    name_from_API,
                    field_type,
                    explicit_layout,
                    offset,
                    field_row_major)
            else:
                yield block_member(
                    name_from_shader,
                    name_from_API,
                    field_type,
                    "",
                    offset,
                    False)
        elif isstructure(field_type):
            yield block_member(
                name_from_shader,
                name_from_API,
                field_type,
                explicit_layout,
                offset,
                field_row_major)

            a = packing.base_alignment(field_type, field_row_major)

            for x in iterate_all_recursive(struct_types[field_type],
                                           None,
                                           name_from_API,
                                           name_from_shader,
                                           packing,
                                           align(offset, a),
                                           field_row_major):
                yield x

        elif ismatrix(field_type):
            yield block_member(
                name_from_shader,
                name_from_API,
                field_type,
                explicit_layout,
                offset,
                field_row_major)
        elif isvector(field_type) or isscalar(field_type):
            yield block_member(
                name_from_shader,
                name_from_API,
                field_type,
                "",
                offset,
                False)
        else:
            raise Exception("Malformed type name {}".format(field_type))

        a = packing.base_alignment(field_type, field_row_major)
        offset = align(offset, a) + packing.size(field_type, field_row_major)


def iterate_all_block_members(fields,
                              field_layouts,
                              block_name,
                              instance_name,
                              packing,
                              row_major):
    """In-order recursive iteration of all subfields of a block

    fields -- List of fields in the uniform block
    field_layouts -- Layout for each field in the uniform block
    block_name - Name of the uniform block
    instance_name - Instance name associated with the block.  May be None.
    packing -- Packing rules used to layout the type
    row_major -- True if the default block matrix orientation is row-major

    At each step of the iteration will yield a block_member object
    representing the member.

    Basic GLSL types will be visited once.  Arrays of basic GLSL types will
    also be visited once.

    A structure will be visited once, then each field in the structure will be
    visited.

    An array of structures will be visited once, then each element of the
    array will be visited.  Since each element is a structure, each field will
    be visited.

    Consider the following block:

        struct S {
            int i;
            float f[2];
        };

        uniform U {
            S s[2];
        };

    The members would be visited in the following order: s, s[0], s[0].i,
    s[0].f[0], s[1], s[1].i, s[1].f[0].
    """
    for x in iterate_all_recursive(fields,
                                   field_layouts,
                                   block_name,
                                   instance_name,
                                   packing,
                                   0,
                                   row_major):
        yield x


def hash_string(string):
    """The djb2 string hash algorithm from the old comp.lang.c days.  Not a
    terrific hash, but we just need a pseudorandom number based on the string.
    This will do.

    This is used instead of the built-in hash function so that the results
    will be consistent on other systems.  See
    http://stackoverflow.com/questions/793761/built-in-python-hash-function.
    """
    h = 5381

    for c in string:
        h = h * 33 + ord(c)

    return h & 0x0ffffffff


def random_data(type, name, offset):
    """Generate pseudorandom data.

    The data generated is based on the type, name of the field, and offset of
    the member in the UBO.
    """
    if isscalar(type):
        h = hash_string("{}@{}".format(offset, name))

        if type == "int":
            return str(h - 0x7fffffff)
        elif type == "uint":
            return str(h)
        elif type == "bool":
            return str(int((h & 8) == 0))
        elif type == "float" or type == "double":
            return str(float(h - 0x7fffffff) / 65535.0)
        else:
            raise Exception("Unknown scalar type {}".format(type))

    if isvector(type):
        scalar = vector_base_type(type)

        x = [random_data(scalar, name, offset + (i * 3))
             for i in xrange(vector_size(type))]
        return " ".join(x)

    if ismatrix(type):
        r, c = matrix_dimensions(type)

        x = [random_data("float", name, offset + (i * 7))
             for i in xrange(r * c)]
        return " ".join(x)


def generate_test_vectors(fields,
                          field_layouts,
                          block_name,
                          instance_name,
                          packing,
                          row_major):
    test_vectors = []

    for m in iterate_all_block_members(fields,
                                       field_layouts,
                                       block_name,
                                       instance_name,
                                       packing,
                                       row_major):
        a = packing.base_alignment(m.GLSL_type, m.row_major)

        if isarray(m.GLSL_type):
            size = array_elements(m.GLSL_type)
            base_type = array_base_type(m.GLSL_type)
            astride = packing.array_stride(m.GLSL_type, m.row_major)
            name = m.API_name + "[0]"
            while(isarray(base_type)):
                size = array_elements(base_type)
                base_type = array_base_type(base_type)
                name = name + "[0]"
        else:
            base_type = m.GLSL_type
            astride = 0
            name = m.API_name
            size = m.size

        if ismatrix(base_type):
            test_vectors.append((
                    name,
                    m.API_type,
                    size,
                    align(m.offset, a),
                    astride,
                    packing.matrix_stride(base_type, m.row_major),
                    int(m.row_major)))
        elif isvector(base_type) or isscalar(base_type):
            test_vectors.append((
                    name,
                    m.API_type,
                    size,
                    align(m.offset, a),
                    astride,
                    0,
                    0))

    return test_vectors


def scalar_derp(type, name, data):
    """Return a GLSL code string to compare a scalar with its expected value."""
    if type == "bool":
        if int(data) == 0:
            return name
        else:
            return "!" + name
    elif type == "uint":
        return "{} != {}u".format(name, data)
    elif type == "int":
        return "{} != {}".format(name, data)
    elif type == "float":
        # Not all implementations support the bit-cast operators that are used
        # to do bit-exact comparisons.  For this reason float_match needs the
        # float data and the bit-exact data.

        bits = bit_exact_data(data, "float")
        return "!float_match({}, {}, {}u)".format(name, data, bits)
    elif type == "double":
        bits = bit_exact_data(data, "double")

        # 0xHHHHHHHHLLLLLLLL
        # 012345678901234567

        hi = "0x" + bits[2:10]
        lo = "0x" + bits[10:18]

        return "!double_match({}, uvec2({}, {}))".format(name, lo, hi)
    else:
        raise Exception("Unknown scalar type {}".format(type))


def vector_derp(type, name, data):
    """Return a list of GLSL code strings that compare each field of a vector
    its expected value.
    """
    scalar = vector_base_type(type)
    components = ["x", "y", "z", "w"]

    return [scalar_derp(scalar,
                 "{}.{}".format(name, "xyzw"[i]),
                 data[i])
            for i in xrange(vector_size(type))]


def matrix_derp(type, name, data):
    """Return a list of GLSL code strings that compare each field of a matrix
    its expected value.
    """
    c, r = matrix_dimensions(type)

    if type[0] == 'd':
        column_type = "dvec{}".format(r)
    else:
        column_type = "vec{}".format(r)

    data_pairs = []

    for i in xrange(c):
        data_pairs.extend(vector_derp(
                column_type,
                "{}[{}]".format(name, i),
                data[(i * r):(i * r) + r]))

    return data_pairs


def fudge_type_for_setter(type):
    """Return the correct type for the GL API to set the uniform.

    Most GLSL data types are set using a the enum name for the actual type.
    The sole exception is bool and related vector types.  For bool types, the
    same-sized (scalar or vector) integer type is used.
    """
    if type[0] == 'b':
        if type == "bool":
            return "int"
        else:
            return "i" + type[1:]
    else:
        return type


def bit_exact_data(raw_data, type):
    """Several places in the test want bit-exact data for all types.
    For integer types, this is just the raw data.  For floating point (both
    single and double precission) types, the bit-exact data is the hex
    representation of the IEEE 754 encoding.
    """
    if type in ["float", "vec2",   "vec3",   "vec4",
                "mat2",  "mat2x2", "mat2x3", "mat2x4",
                "mat3",  "mat3x2", "mat3x3", "mat3x4",
                "mat4",  "mat4x2", "mat4x3", "mat4x4"]:
        exact_data = []

        for d in raw_data.split(" "):
            p = struct.pack('!f', float(d))
            u = struct.unpack('!I', p)[0]
            exact_data.append(hex(u))

        return " ".join(exact_data)
    elif type in ["double", "dvec2",   "dvec3",   "dvec4",
                  "dmat2",  "dmat2x2", "dmat2x3", "dmat2x4",
                  "dmat3",  "dmat3x2", "dmat3x3", "dmat3x4",
                  "dmat4",  "dmat4x2", "dmat4x3", "dmat4x4"]:
        exact_data = []

        for d in raw_data.split(" "):
            p = struct.pack('!d', float(d))
            u = struct.unpack('!Q', p)[0]

            # Sometimes the hex() generates a spurious "L" at the end of the
            # string.  Only take the first 18 characters to omit the unwanted
            # "L".  I believe this occurs when bit 63 is set.

            exact_data.append(hex(u)[0:18])

        return " ".join(exact_data)
    else:
        return raw_data

def generate_array_data_pairs(name, api_name, element_type, row_major, offset, packing, setters, checkers):
    base_type = array_base_type(element_type)

    astride = packing.array_stride(element_type, row_major)

    for i in xrange(array_elements(element_type)):

        name_with_index = "{}[{}]".format(name, i)
        api_name_with_index = "{}[{}]".format(api_name, i)

        if isarray(base_type):
            offset = offset + (i * arrays_of_arrays_size(base_type) * astride)
            generate_array_data_pairs(name_with_index, api_name_with_index, base_type, row_major, offset, packing, setters, checkers)
        else:
            offset = offset + (i * astride)
            raw_data = random_data(base_type, name_with_index, offset)
            setters.append(
                (fudge_type_for_setter(base_type),
                api_name_with_index,
                bit_exact_data(raw_data, base_type)))

            data = raw_data.split(" ")

            if isscalar(base_type):
                checkers.append(scalar_derp(base_type,
                                            name_with_index,
                                            data[0]))
            elif isvector(base_type):
                checkers.extend(vector_derp(base_type,
                                            name_with_index,
                                            data))
            elif ismatrix(base_type):
                checkers.extend(matrix_derp(base_type,
                                            name_with_index,
                                            data))

def generate_data_pairs(uniform_blocks, packing):
    """Return GLSL code to test values and shader_runner code to set them.

    For each element in each uniform block, two lists are generated, and these
    lists are returned in a tuple of (checkers, setters).  The "checkers" list
    is a list of GLSL comparisons that return true if the contents of the
    uniform block element DOES NOT match the expected value.

    The "setters" list is a list of shader_runner "uniform" statements to set
    each uniform block member to an expected value.
    """
    checkers = []
    setters = []

    for (block_name,
         instance_name,
         global_layout,
         block_layout,
         fields,
         field_layouts) in uniform_blocks:
        for m in iterate_all_block_members(
            fields,
            field_layouts,
            block_name,
            instance_name,
            packing,
            block_row_major_default(global_layout, block_layout)):

            if m.API_type:
                if isarray(m.GLSL_type):
                    generate_array_data_pairs(m.GLSL_name, m.API_name, m.GLSL_type, m.row_major, m.offset, packing, setters, checkers)
                else:
                    raw_data = random_data(m.GLSL_type, m.GLSL_name, m.offset)
                    setters.append((fudge_type_for_setter(m.GLSL_type),
                                    m.API_name,
                                    bit_exact_data(raw_data, m.GLSL_type)))

                    data = raw_data.split(" ")

                    if isscalar(m.GLSL_type):
                        checkers.append(scalar_derp(m.GLSL_type,
                                                    m.GLSL_name,
                                                    data[0]))
                    elif isvector(m.GLSL_type):
                        checkers.extend(vector_derp(m.GLSL_type,
                                                    m.GLSL_name,
                                                    data))
                    elif ismatrix(m.GLSL_type):
                        checkers.extend(matrix_derp(m.GLSL_type,
                                                    m.GLSL_name,
                                                    data))

    return checkers, setters


def pretty_format_type_data(packing, type, offset, row_major):
    """Return a string describing the layout of the type

    This string is intended to be a comment embedded in the GLSL shader code
    that is generated.
    """
    a = packing.base_alignment(type, row_major)
    aligned_offset = align(offset, a)
    size = packing.size(type, row_major)

    row_major_str = "-"
    mstride = "-"
    astride = "-"

    if isarray(type):
        astride = packing.array_stride(type, row_major)

        base_type = without_array(type)
        if ismatrix(base_type) and row_major:
            if row_major:
                row_major_str = "yes"
            else:
                row_major_str = "no"

            mstride = packing.matrix_stride(base_type, row_major)
    else:
        if ismatrix(type):
            if row_major:
                row_major_str = "yes"
            else:
                row_major_str = "no"

            mstride = packing.matrix_stride(type, row_major)

    return "{base_align:>3}  {base_offset:>4}  {aligned_offset:>5}  {padded_size:>6}  {row_major:^5}  {array_stride:>6}  {matrix_stride:>6}".format(
        base_align=a,
        base_offset=offset,
        aligned_offset=aligned_offset,
        padded_size=size,
        row_major=row_major_str,
        array_stride=astride,
        matrix_stride=mstride
        )


def pretty_format_member(m, packing):
    """Return a string of GLSL code to declare the member.

    m -- a block_member object
    packing -- packing rules

    If the member is "nested" inside a structure, the string returned will be
    a comment simply describing the layout of the submember.
    """
    # If the name ends in an array subscript, emit a special line to note that
    # the following fields are the contents of an element of an array of
    # structures.
    if m.GLSL_name[-1] == "]":
        n = m.struct_nesting() + 1
        indent = "//  " + ("  " * n)

        return "{indent}{index}".format(indent=indent,
                                         index=m.GLSL_name[m.GLSL_name.index("["):])

    # Strip off everything before the last period.
    name = m.GLSL_name.split(".")[-1]

    n = m.struct_nesting()
    if n > 0:
        indent = "//  " + ("  " * n)
        field_str = "{indent}{type:<11} {name:<20}".format(
            indent=indent,
            type=m.GLSL_type,
            name=name)[0:31]
    else:
        field_str = "    {type:<11}{name};{padding}//   ".format(
            type=m.GLSL_type,
            name=name,
            padding="          "[len(name):])

    data_str = pretty_format_type_data(
        packing,
        m.GLSL_type,
        m.offset,
        m.row_major)

    # If there is an explicit layout for the member, prepend it to the member
    # declaration.  This also means that the member must be contained directly
    # in the UBO (i.e., not nested in a struct), so no additional indentation
    # is necessary.

    if m.explicit_layout and "#" not in m.explicit_layout:
        return "    layout({layout})\n{field}{data}".format(
            layout=m.explicit_layout,
            field=field_str,
            data=data_str)
    else:
        return "{field}{data}".format(field=field_str, data=data_str)


def block_row_major_default(global_layout, block_layout):
    """Return true if the default matrix layout for the block is row-major."""
    row_major = global_layout is not None and "row_major" in global_layout

    if block_layout:
        if "row_major" in block_layout:
            row_major = True
        elif "column_major" in block_layout:
            # The block layout can override a previous global layout.
            row_major = False

    return row_major


def generate_block_list(glsl_version, packing, ubo_fields, layouts):
    """Return list of uniform blocks to be passed to emit_shader_test.

    The supplied ubo_fields and layouts are reformatted as an element in a
    list of blocks that can be consumed by emit_shader_test.  Depending on the
    value of glsl_version, additional blocks may be added to the list.  The
    new blocks will have the same layout as the original, but the layout will
    be specified in different ways (e.g., by using a global layout qualifier).
    """

    blocks = [("UB1", "", None, packing.layout_string(), ubo_fields, layouts)]

    # If the GLSL version is at least 1.50, UBO functionality is significantly
    # extended.
    #
    # 1. Uniform blocks can have instance names.  The existence of the name
    #    changes the way the block is queried through the API (with the block
    #    name) and the way it is accessed by the shader (with the instance
    #    name).
    #
    # 2. Uniform blocks can be grouped in arrays.  UBO arrays must have an
    #    instance name.
    #
    # This is used to make the tests dramatically more complex.  Each UBO is
    # emitted three times.
    #
    # 1. Without an instance name.
    #
    # 2. With an instance name and the per-block matrix layout switched to
    #    row_major.  The declared layout of the individual fields is modified
    #    so that this block has the same layout as the previous block.
    #
    # 3. With an instance name and an array size.  The per-block matrix layout
    #    is empty, but the global matrix layout is changed to row_major.  This
    #    block should have the same layout as the previous two.

    if glsl_version >= 150:
        inverted_layouts = [relayout_for_default_row_major(l) for l in layouts]

        blocks.append(("UB2",
                       "ub2",
                       None,
                       packing.layout_string() + ", row_major",
                       ubo_fields,
                       inverted_layouts))

        blocks.append(("UB3",
                       "ub3",
        # Disabled to work around Mesa bug #83508.
        #               "ub3[2]",
                       packing.layout_string() + ", row_major",
                       None,
                       ubo_fields,
                       inverted_layouts))

    return blocks


def emit_shader_test(blocks, packing, glsl_version, extensions):
    structures = []
    test_vectors = []

    for (block_name,
         instance_name,
         global_layout,
         block_layout,
         fields,
         field_layouts) in blocks:

        structures.extend([s for s in iterate_structures(fields)])

        test_vectors.extend(generate_test_vectors(
                fields,
                field_layouts,
                block_name,
                instance_name,
                packing,
                block_row_major_default(global_layout, block_layout)))


    checkers, setters = generate_data_pairs(blocks, packing)

    # If the GLSL version is at least 1.40, UBOs are already supported, and we
    # don't need to enable the extension.

    if glsl_version >= 140 and "GL_ARB_uniform_buffer_object" in extensions:
        extensions.remove("GL_ARB_uniform_buffer_object")

    t = Template(dedent("""\
    [require]
    GLSL >= ${glsl_version / 100}.${glsl_version % 100}
    % for ext in extensions:
    ${ext}
    % endfor

    # Do NOT edit the following lines.
    # GLSL ${glsl_version}
    # EXTENSIONS ${extensions}
    # PACKING ${packing.layout_string()}
    % for s in structures:
    # STRUCT ("${s}", ${struct_types[s]})
    % endfor
    % for b in uniform_blocks:
    # UBO ${b}
    % endfor
    # DATA END

    [vertex shader]
    % for ext in extensions:
    #extension ${ext}: require
    % endfor
    #extension GL_ARB_shader_bit_encoding: enable
    #extension GL_ARB_gpu_shader5: enable

    precision highp float;
    % for s in structures:

    struct ${s} {
        % for (field_type, field_name) in struct_types[s]:
        ${"{:<11}".format(field_type)} ${field_name};
        % endfor
    };
    % endfor

    % for (block_name, instance_name, global_layout, block_layout, fields, field_layouts) in uniform_blocks:
    % if global_layout:
    layout(${global_layout}) uniform;

    % endif
    % if block_layout:
    layout(${block_layout})
    % endif
    uniform ${block_name} {
                              // base   base  align  padded  row-   array   matrix
                              // align  off.  off.   size    major  stride  stride
    % for m in iterate_all_block_members(fields, field_layouts, block_name, instance_name, packing, block_row_major_default(global_layout, block_layout)):
    ${pretty_format_member(m, packing)}
    % endfor
    } ${instance_name};
    % endfor

    flat out int vertex_pass;
    in vec4 piglit_vertex;

    #if defined(GL_ARB_shader_bit_encoding) || defined(GL_ARB_gpu_shader5) || __VERSION__ >= 430
    bool float_match(float u, float f, uint bits) { return floatBitsToUint(u) == bits; }
    #else
    bool float_match(float u, float f, uint bits) { return u == f; }
    #endif
    % if glsl_version >= 400 or "GL_ARB_gpu_shader_fp64" in extensions:

    bool double_match(double u, uvec2 bits) { return unpackDouble2x32(u) == bits; }
    %endif

    void main()
    {
        /* std140 (or shared) layout prevents any fields or blocks from being
         * eliminated.  Section 2.11.6 of the OpenGL ES 3.0 spec makes this
         * explicit, but desktop GL specs only imply it.
         */
        bool pass = true;

    % for i in xrange(len(checkers)):
        % if i % 5 == 0:
        if (${checkers[i]})
            pass = false;
        % endif
    % endfor

        vertex_pass = int(pass);
        gl_Position = piglit_vertex;
    }

    [fragment shader]
    precision highp float;

    out vec4 piglit_fragcolor;
    flat in int vertex_pass;

    void main()
    {
        piglit_fragcolor = bool(vertex_pass) ? vec4(0, 1, 0, 1) : vec4(1, 0, 0, 1);
    }

    [test]
    link success
    % for (name, type, size, offset, astride, mstride, row_major) in test_vectors:

    active uniform ${name} GL_UNIFORM_TYPE ${type}
    active uniform ${name} GL_UNIFORM_SIZE ${size}
        % if packing.fixed_offsets():
    active uniform ${name} GL_UNIFORM_OFFSET ${offset}
    active uniform ${name} GL_UNIFORM_ARRAY_STRIDE ${astride}
    active uniform ${name} GL_UNIFORM_MATRIX_STRIDE ${mstride}
        % endif
    active uniform ${name} GL_UNIFORM_IS_ROW_MAJOR ${row_major}
    % endfor

    % for (type, name, data) in setters:
    uniform ${type} ${name} ${data}
    % endfor

    draw rect -1 -1 2 2
    probe all rgba 0.0 1.0 0.0 1.0"""))

    return t.render(glsl_version=glsl_version,
                    extensions=extensions,
                    structures=structures,
                    test_vectors=test_vectors,
                    uniform_blocks=blocks,
                    packing=packing,
                    iterate_all_block_members=iterate_all_block_members,
                    pretty_format_member=pretty_format_member,
                    block_row_major_default=block_row_major_default,
                    struct_types=struct_types,
                    checkers=checkers,
                    setters=setters)


def generate_file_name(requirements, packing):
    """Based on the required UBO elements and the packing, generate a
    plausible file name for the test case.
    """
    prefix = packing.layout_string() + "-"
    suffix = ".shader_test"

    body = "-and-".join(["-".join(req) for req in requirements])

    return prefix + body + suffix


class packing_rules(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def layout_string(self):
        """Get the string used in a layout qualifier to select this set of
        layout rules.
        """

    @abc.abstractproperty
    def fixed_offsets(self):
        """Return true if fields in this layout have fixed locations (e.g.,
        std140) or false if they can vary among implementations (e.g., shared
        or packed)?
        """

    @abc.abstractmethod
    def base_alignment(self, type, row_major):
        """Determine the base alignment, in bytes, of the named type"""

    @abc.abstractmethod
    def matrix_stride(self, type, row_major):
        """Determine the stride, in bytes, from one indexable vector of the
        matrix (column or row depending on the orientation) to the next.
        """

    @abc.abstractmethod
    def array_stride(self, type, row_major):
        """Determine the stride, in bytes, from one array element to the next.
        If the type is not an array type, zero is returned.
        """

    def size(self, type, row_major):
        """Determine the size, in bytes, of the specified type.

        Applying the packing rules and the row-major setting, determine the
        size, in bytes, of the entire data type.  This will include padded
        after the data element as required by the GLSL specification.
        """
        if isarray(type):
            return self.array_stride(type, row_major) * arrays_of_arrays_size(type)

        if type in ["float", "bool", "int", "uint"]:
            return 4

        if type == "double":
            return 8

        if type in ["vec2", "bvec2", "ivec2", "uvec2"]:
            return 2 * 4

        if type == "dvec2":
            return 2 * 8

        if type in ["vec3", "bvec3", "ivec3", "uvec3"]:
            return 3 * 4

        if type == "dvec3":
            return 3 * 8

        if type in ["vec4", "bvec4", "ivec4", "uvec4"]:
            return 4 * 4

        if type == "dvec4":
            return 4 * 8

        if "mat" in type:
            c, r = matrix_dimensions(type)
            if not row_major:
                return c * self.matrix_stride(type, row_major)
            else:
                return r * self.matrix_stride(type, row_major)

        if type not in struct_types:
            raise Exception("Unknown type {}".format(type))

        s = 0
        fields = struct_types[type]
        for (t, n) in fields:
            a = self.base_alignment(t, row_major)

            s = align(s, a) + self.size(t, row_major)

        s = align(s, self.base_alignment(type, row_major))
        return s


class std140_packing_rules(packing_rules):
    def layout_string(self):
        return "std140"

    def fixed_offsets(self):
        return True

    def base_alignment(self, type, row_major):
        # (4) If the member is an array of scalars or vectors, the base
        #     alignment and array stride are set to match the base alignment
        #     of a single array element, according to rules (1), (2), and (3),
        #     and rounded up to the base alignment of a vec4. The array may
        #     have padding at the end; the base offset of the member following
        #     the array is rounded up to the next multiple of the base
        #     alignment.

        if isarray(type):
            return max(16,
                       self.base_alignment(array_base_type(type), row_major))

        # (1) If the member is a scalar consuming <N> basic machine units, the
        #     base alignment is <N>.

        if isscalar(type):
            return basic_machine_units(type)

        if isvector(type):
            # (2) If the member is a two- or four-component vector with
            #     components consuming <N> basic machine units, the base
            #     alignment is 2<N> or 4<N>, respectively.
            #
            # (3) If the member is a three-component vector with components
            #     consuming <N> basic machine units, the base alignment is
            #     4<N>.

            components = vector_size(type)
            if components == 2 or components == 4:
                return components * basic_machine_units(vector_base_type(type))
            elif components == 3:
                return 4 * basic_machine_units(vector_base_type(type))

            raise Exception("Invalid vector size {} for type {}".format(
                components,
                type))
        elif ismatrix(type):
            return self.matrix_stride(type, row_major)

        if type not in struct_types:
            raise Exception("Unknown type {}".format(type))

        # (9) If the member is a structure, the base alignment of the
        #     structure is <N>, where <N> is the largest base alignment value
        #     of any of its members, and rounded up to the base alignment of a
        #     vec4. The individual members of this sub-structure are then
        #     assigned offsets by applying this set of rules recursively,
        #     where the base offset of the first member of the sub-structure
        #     is equal to the aligned offset of the structure. The structure
        #     may have padding at the end; the base offset of the member
        #     following the sub-structure is rounded up to the next multiple
        #     of the base alignment of the structure.

        a = 16
        fields = struct_types[type]
        for (field_type, field_name) in fields:
            a = max(a, self.base_alignment(field_type, row_major))

        return a

    def matrix_stride(self, type, row_major):
        c, r = matrix_dimensions(type)
        if not row_major:
            # (4) If the member is an array of scalars or vectors, the base
            #     alignment and array stride are set to match the base
            #     alignment of a single array element, according to rules (1),
            #     (2), and (3), and rounded up to the base alignment of a
            #     vec4. The array may have padding at the end; the base offset
            #     of the member following the array is rounded up to the next
            #     multiple of the base alignment.
            #
            # (5) If the member is a column-major matrix with <C> columns and
            #     <R> rows, the matrix is stored identically to an array of
            #     <C> column vectors with <R> components each, according to
            #     rule (4).

            if type[0] == 'd':
                return max(16, self.base_alignment("dvec{}".format(r), False))
            else:
                return max(16, self.base_alignment("vec{}".format(r), False))
        else:
            # (7) If the member is a row-major matrix with <C> columns and <R>
            #     rows, the matrix is stored identically to an array of <R>
            #     row vectors with <C> components each, according to rule (4).

            if type[0] == 'd':
                return max(16, self.base_alignment("dvec{}".format(c), False))
            else:
                return max(16, self.base_alignment("vec{}".format(c), False))

    def array_stride(self, type, row_major):
        base_type = without_array(type)

        if not isstructure(base_type):
            # (4) If the member is an array of scalars or vectors, the base
            #     alignment and array stride are set to match the base
            #     alignment of a single array element, according to rules (1),
            #     (2), and (3), and rounded up to the base alignment of a
            #     vec4. The array may have padding at the end; the base offset
            #     of the member following the array is rounded up to the next
            #     multiple of the base alignment.
            return max(16,
                       max(self.base_alignment(base_type, row_major),
                           self.size(base_type, row_major)))
        else:
            # (9) If the member is a structure, the base alignment of the
            #     structure is <N>, where <N> is the largest base alignment
            #     value of any of its members, and rounded up to the base
            #     alignment of a vec4. The individual members of this
            #     sub-structure are then assigned offsets by applying this set
            #     of rules recursively, where the base offset of the first
            #     member of the sub-structure is equal to the aligned offset
            #     of the structure. The structure may have padding at the end;
            #     the base offset of the member following the sub-structure is
            #     rounded up to the next multiple of the base alignment of the
            #     structure.
            #
            # (10) If the member is an array of <S> structures, the <S> elements
            #     of the array are laid out in order, according to rule (9).

            return align(self.size(base_type, row_major),
                         self.base_alignment(base_type, row_major))


class shared_packing_rules(std140_packing_rules):
    def layout_string(self):
        return "shared"

    def fixed_offsets(self):
        return False


class unique_name_dict:
    """Helper class to generate a unique name for each field.

    Fields for each particular type have names based on the type.  Each name
    is suffixed by a number, and the number is incremented each time a field
    of that type is created.  For example, the first int field will be named
    "i1", and the second will be named "i2".
    """

    def __init__(self):
        self.names = {}

    def trim_name(self, type):
        """Internal method to reduce a type name to a canonical form

        These names are used to track the type in the self.names dict.

        Array types have the arrayness stripped.  Vector types have vector
        count stripped.  Matrix types are converted to canonical "mat#x#"
        form.  All other type names are unmodified.
        """
        if isarray(type):
            t = without_array(type)
        else:
            t = type

        if ismatrix(t):
            # Canonicalize matrix type names.
            c, r = matrix_dimensions(t)

            name = "mat{}x{}".format(c, r)
            if t[0] == "d":
                name = "d" + name

            return name
        elif isscalar(t):
            return t
        elif isvector:
            return t.strip("1234")
        else:
            # Assume it must be a structure.
            return t

    def add_type(self, type):
        """Internal method to add a previously unknown type to the dictionary

        type -- Actual GLSL type name

        Generates the base name for fields of this type.  Adds a tuple to dict
        (indexed by "type") of (base_name, 1).
        """
        if isarray(type):
            t = without_array(type)
        else:
            t = type

        if isvector(t):
            base = "{}v".format(vector_base_type(t)[0])
        elif ismatrix(t):
            c, r = matrix_dimensions(t)

            if t[0] == 'd':
                base = "dm{}{}_".format(c, r)
            else:
                base = "m{}{}_".format(c, r)
        elif isscalar(t):
            base = t[0]
        elif t[0] == "S":
            base = "s{}_".format(t[1:])
        else:
            raise Exception("Malformed type name {}".format(t))

        self.names[self.trim_name(t)] = (base, 1)
        return

    def get_name(self, type):
        """Return a unique name for a field of the specified type."""
        t = self.trim_name(type)
        if t not in self.names:
            self.add_type(type)

        base, count = self.names[t]
        self.names[t] = (base, count + 1)

        return "{}{}".format(base, count)


class block_member(object):
    """Helper class to track all of the layout information for a single
    (sub-)member of a uniform block."""

    TYPE_ENUM = {
        'float': "GL_FLOAT",
        'vec2':  "GL_FLOAT_VEC2",
        'vec3':  "GL_FLOAT_VEC3",
        'vec4':  "GL_FLOAT_VEC4",

        'double': "GL_DOUBLE",
        'dvec2':  "GL_DOUBLE_VEC2",
        'dvec3':  "GL_DOUBLE_VEC3",
        'dvec4':  "GL_DOUBLE_VEC4",

        'int':   "GL_INT",
        'ivec2': "GL_INT_VEC2",
        'ivec3': "GL_INT_VEC3",
        'ivec4': "GL_INT_VEC4",

        'uint':  "GL_UNSIGNED_INT",
        'uvec2': "GL_UNSIGNED_INT_VEC2",
        'uvec3': "GL_UNSIGNED_INT_VEC3",
        'uvec4': "GL_UNSIGNED_INT_VEC4",

        'bool':  "GL_BOOL",
        'bvec2': "GL_BOOL_VEC2",
        'bvec3': "GL_BOOL_VEC3",
        'bvec4': "GL_BOOL_VEC4",

        'mat2':   "GL_FLOAT_MAT2",
        'mat2x2': "GL_FLOAT_MAT2",
        'mat2x3': "GL_FLOAT_MAT2x3",
        'mat2x4': "GL_FLOAT_MAT2x4",

        'mat3':   "GL_FLOAT_MAT3",
        'mat3x2': "GL_FLOAT_MAT3x2",
        'mat3x3': "GL_FLOAT_MAT3",
        'mat3x4': "GL_FLOAT_MAT3x4",

        'mat4':   "GL_FLOAT_MAT4",
        'mat4x2': "GL_FLOAT_MAT4x2",
        'mat4x3': "GL_FLOAT_MAT4x3",
        'mat4x4': "GL_FLOAT_MAT4",

        'dmat2':   "GL_DOUBLE_MAT2",
        'dmat2x2': "GL_DOUBLE_MAT2",
        'dmat2x3': "GL_DOUBLE_MAT2x3",
        'dmat2x4': "GL_DOUBLE_MAT2x4",

        'dmat3':   "GL_DOUBLE_MAT3",
        'dmat3x2': "GL_DOUBLE_MAT3x2",
        'dmat3x3': "GL_DOUBLE_MAT3",
        'dmat3x4': "GL_DOUBLE_MAT3x4",

        'dmat4':   "GL_DOUBLE_MAT4",
        'dmat4x2': "GL_DOUBLE_MAT4x2",
        'dmat4x3': "GL_DOUBLE_MAT4x3",
        'dmat4x4': "GL_DOUBLE_MAT4"
    }

    def __init__(self,
                 GLSL_name,
                 API_name,
                 GLSL_type,
                 explicit_layout,
                 offset,
                 row_major):
        self.GLSL_name = GLSL_name
        self.GLSL_type = GLSL_type

        self.API_name = API_name

        self.explicit_layout = explicit_layout
        self.offset = offset
        self.row_major = row_major

        if isarray(GLSL_type):
            base_type = without_array(GLSL_type)
            if isstructure(base_type):
                self.API_type = None
            else:
                self.API_type = self.TYPE_ENUM[base_type];

            self.size = array_elements(GLSL_type)
        elif isstructure(GLSL_type):
            self.API_type = None
            self.size = 1
        else:
            self.API_type = self.TYPE_ENUM[GLSL_type];
            self.size = 1

    def struct_nesting(self):
        if "." in self.GLSL_name:
            # If the block has an instance name, the API name will use the
            # block name instead of the instance name.  As a result,
            # GLSL_name and API_name will be different.
            #
            # The first "." is for the block instance name, so it does not count
            # as structure nesting.

            if self.GLSL_name != self.API_name:
                return collections.Counter(self.GLSL_name)["."] - 1
            else:
                return collections.Counter(self.GLSL_name)["."]
        else:
            return 0


if __name__ == "__main__":
    if len(sys.argv) > 1:
        max_glsl_version = int(sys.argv[1])
    else:
        max_glsl_version = 130

    if len(sys.argv) > 2:
        extensions = sys.argv[2:]
    else:
        extensions = []

    # Pick a random GLSL version from the available set of possible versions.
    glsl_version = random.choice([v for v in [130, 140, 150, 400, 430]
                                  if v <= max_glsl_version])

    # Use the GLSL version filter out some extensions that are redundant.
    if glsl_version >= 140 and "GL_ARB_uniform_buffer_object" in extensions:
        extensions.remove("GL_ARB_uniform_buffer_object")

    if glsl_version >= 400 and "GL_ARB_gpu_shader_fp64" in extensions:
        extensions.remove("GL_ARB_gpu_shader_fp64")

    if glsl_version >= 430 and "GL_ARB_arrays_of_arrays" in extensions:
        extensions.remove("GL_ARB_arrays_of_arrays")

    # Pick a random subset of the remaining extensions.
    num_ext = len(extensions)
    if num_ext > 0:
        random.shuffle(extensions)
        r = random.randint(0, num_ext)
        extensions = extensions[:r]

    # Based on the GLSL version and the set of extensions, pick the set of
    # possible data types.
    if glsl_version < 400:
        types = ALL130_TYPES
    else:
        types = ALL400_TYPES

    if "GL_ARB_gpu_shader_fp64" in extensions:
        types.extend(DOUBLE_TYPES)

    # Based on the GLSL version, pick a set of packing rules
    # FINISHME: Add support for std430_packing_rules() soon.
    packing = random.choice([std140_packing_rules(), shared_packing_rules()])

    # Based on the GLSL version and the set of available extensions, pick
    # some required combinations of data structures to include in the UBO.
    arrays_of_arrays = (glsl_version >= 430 or
                        "GL_ARB_arrays_of_arrays" in extensions)

    allow_row_major_structure = glsl_version >= 150

    requirements = []
    for i in [1, 2]:
        x = [random.choice(["array", "struct"])]

        for j in [1, 2, 3]:
            # If arrays-of-arrays are not supported, don't allow "array" to be
            # picked twice in a row.

            if x[-1] == "array" and not arrays_of_arrays:
                x.append("struct")
            else:
                x.append(random.choice(["array", "struct"]))

        if "struct" in x and allow_row_major_structure:
            ordering = random.choice([None,
                                      None,
                                      None,
                                      None,
                                      "column_major",
                                      "#column_major",
                                      "row_major",
                                      "row_major"])
            if ordering:
                x = [ordering] + x

        requirements.append(x)

    if glsl_version < 140:
        extensions.append("GL_ARB_uniform_buffer_object")

    # Generate the test!
    fields, required_layouts = generate_ubo(requirements, types)

    layouts = generate_layouts(
        fields,
        required_layouts,
        allow_row_major_structure)

    blocks = generate_block_list(
        glsl_version,
        packing,
        fields,
        layouts)

    print(emit_shader_test(
        blocks,
        packing,
        glsl_version,
        extensions))



# coding=utf-8
#
# Copyright © 2016 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Generate fp64 and int64 types conversion tests."""

from __future__ import print_function, division, absolute_import
import abc
import argparse
import itertools
import os
import struct

import numpy as np

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

# pylint: disable=bad-whitespace,line-too-long,bad-continuation
DOUBLE_INFS                 = ['0xfff0000000000000', # -inf
                               '0x7ff0000000000000'] # +inf

DOUBLE_NEG_ZERO             = ['0x8000000000000000'] # Negative underflow (-0.0)

DOUBLE_POS_ZERO             = ['0x0000000000000000'] # Positive underflow (+0.0)

# Double values causing an underflow to zero in any other type
DOUBLE_DENORMAL_VALUES      = ['0x800fffffffffffff', # Negative maximum denormalized -- Denormalized may be flushed to 0
                               '0x8000000000000001', # Negative minimum denormalized -- Denormalized may be flushed to 0
                               '0x0000000000000001', # Positive minimum denormalized -- Denormalized may be flushed to 0
                               '0x000fffffffffffff'] # Positive maximum denormalized -- Denormalized may be flushed to 0

DOUBLE_NORMAL_VALUES        = ['0x8010000000000000', # Negative minimum normalized
                               '0x0010000000000000'] # Positive minimum normalized

# Double +/-inf
DOUBLE_FLOAT_INFS           = ['0xc7effffff0000000', # Negative overflow (-inf)
                               '0x47effffff0000000'] # Positive overflow (+inf)

DOUBLE_FLOAT_VALUES         = ['0xc7efffffefffffff', # Negative maximum normalized
                               '0xc170000000000000', # -16777216.0
                               '0xc014000000000000', # -5.0
                               '0xbfff25ce60000000', # -1.9467300176620483
                               '0xb80fffffe0000000', # Negative minimum normalized
                               '0xb69fffffffffffff', # Negative underflow
                               '0x369fffffffffffff', # Positive underflow
                               '0x380fffffe0000000', # Positive minimum normalized
                               '0x3fff25ce60000000', # +1.9467300176620483
                               '0x4014000000000000', # +5.0
                               '0x4170000000000000', # +16777216.0
                               '0x47efffffefffffff'] # Positive maximum normalized

DOUBLE_UINT_VALUES          = ['0xbfeccccccccccccd', # -0.9
                               #'0x8010000000000000', # Negative minimum normalized          -- Already checked
                               #'0x800fffffffffffff', # Negative maximum denormalized        -- Already checked
                               #'0x8000000000000001', # Negative minimum denormalized        -- Already checked
                               #'0x8000000000000000', # Negative minimum        (-0)         -- Already checked
                               #'0x0000000000000000', # Positive minimum        (+0)         -- Already checked
                               '0x3fff25ce60000000', # +1.9467300176620483
                               '0x4014000000000000', # +5.0
                               '0x4170000000000000', # +16777216.0
                               '0x41dfffffffc00000', # Signed int low frontier (+2147483647)
                               '0x41e0000000000000', # Signed int up frontier  (+2147483648)
                               '0x41efffffffe00000'] # Maximum                 (+4294967295)

DOUBLE_INT_VALUES           = ['0xc1e0000000000000', # Minimum          (-2147483648)
                               '0xc170000000000000', # -16777216.0
                               '0xc014000000000000', # -5.0
                               '0xbfff25ce60000000', # -1.9467300176620483
                               #'0x8000000000000000', # Negative minimum        (-0)         -- Already checked
                               #'0x0000000000000000', # Minimum                 (+0)         -- Already checked
                               '0x3fff25ce60000000', # +1.9467300176620483
                               '0x4014000000000000', # +5.0
                               '0x4170000000000000', # +16777216.0
                               '0x41dfffffffc00000'] # Maximum          (+2147483647)

DOUBLE_BOOL_VALUES          = [#'0x8010000000000000', # Minimum negative True value -- Already checked
                               #'0x0000000000000000', # False                       -- Already checked
                               #'0x0010000000000000', # Minimum positive True value -- Already checked
                              ]

FLOAT_INFS                  = ['0xff800000', # -inf
                               '0x7f800000'] # +inf

FLOAT_NEG_ZERO              = ['0x80000000'] # Negative underflow (-0.0)

FLOAT_POS_ZERO              = ['0x00000000'] # Positive underflow (+0.0)

FLOAT_VALUES                = ['0xff7fffff', # Negative maximum normalized
                               '0xcb800000', # -16777216.0
                               '0xc0a00000', # -5.0
                               '0xbff92e73', # -1.9467300176620483
                               '0x80800000', # Negative minimum normalized
                               #'0x807fffff', # Negative maximum denormalized -- Denormalized may be flushed to 0
                               #'0x80000001', # Negative minimum denormalized -- Denormalized may be flushed to 0
                               #'0x00000001', # Positive minimum denormalized -- Denormalized may be flushed to 0
                               #'0x007fffff', # Positive maximum denormalized -- Denormalized may be flushed to 0
                               '0x00800000', # Positive minimum normalized
                               '0x3ff92e73', # +1.9467300176620483
                               '0x40a00000', # +5.0
                               '0x4b800000', # +16777216.0
                               '0x7f7fffff'] # Positive maximum normalized

FLOAT_INT64_VALUES          = ['0xcf800000', # -4294967296.0
                               '0xcb800000', # -16777216.0
                               '0xc0a00000', # -5.0
                               '0xbff92e73', # -1.9467300176620483
                               '0x80800000', # Negative minimum normalized
                               '0x807fffff', # Negative maximum denormalized
                               '0x80000001', # Negative minimum denormalized
                               '0x00000001', # Positive minimum denormalized
                               '0x007fffff', # Positive maximum denormalized
                               '0x00800000', # Positive minimum normalized
                               '0x3ff92e73', # +1.9467300176620483
                               '0x40a00000', # +5.0
                               '0x4b800000', # +16777216.0
                               '0x4f800000'] # +4294967296.0

FLOAT_UINT64_VALUES         = ['0x00000001', # Positive minimum denormalized
                               '0x007fffff', # Positive maximum denormalized
                               '0x00800000', # Positive minimum normalized
                               '0x3ff92e73', # +1.9467300176620483
                               '0x40a00000', # +5.0
                               '0x4b800000', # +16777216.0
                               '0x4f800000'] # +4294967296.0

UINT64_VALUES               = ['0', # Minimum
                               '5',
                               '2147483647', # Signed int32 low frontier
                               '2147483648', # Signed int32 up frontier
                               '4294967295', # Maximum unsigned int32
                               '4294967296',
                               '9223372036854775807', # Signed int64 low frontier
                               '9223372036854775808', # Signed int64 up frontier
                               '18446744073709551615'] # Maximum

INT64_VALUES                = ['-9223372036854775808', # Minimum
                               '-2147483649',
                               '-2147483648',
                               '-5',
                               '-1',
                               '0',
                               '1',
                               '5',
                               '2147483647', # Signed int32 low frontier
                               '2147483648', # Signed int32 up frontier
                               '4294967295', # Maximum unsigned int32
                               '4294967296',
                               '9223372036854775807'] # Maximum

UINT_VALUES                 = ['0', # Minimum
                               '5',
                               '2147483647', # Signed int low frontier
                               '2147483648', # Signed int up frontier
                               '4294967295'] # Maximum

INT_VALUES                  = ['-2147483648', # Minimum
                               '-5',
                               '-1',
                               '0',
                               '1',
                               '5',
                               '2147483647'] # Maximum

BOOL_VALUES                 = ['0', # False
                               '1'] # True
# pylint: enable=bad-whitespace,line-too-long,bad-continuation

def get_dir_name(ver, test_type):
    """Returns the directory name to save tests given a GLSL version and a
       test type.
    """

    assert isinstance(ver, str)
    assert isinstance(test_type, str)
    if ver.startswith('GL_'):
        feature_dir = ver[3:].lower()
    else:
        feature_dir = 'glsl-{}.{}'.format(ver[0], ver[1:])

    return os.path.join('spec', feature_dir, test_type,
                        'conversion')


class TestTuple(object):
    """A float64 derived and other type derived tuple to generate the
       needed conversion tests.
    """

    @staticmethod
    def float_to_hex(fvalue):
        """Returns the hexadecimal representation from a float32 value."""
        assert isinstance(fvalue, np.float32)
        return hex(struct.unpack('<I', struct.pack('<f', fvalue))[0])

    @staticmethod
    def double_to_hex(fvalue):
        """Returns the hexadecimal representation from a float64 value."""
        assert isinstance(fvalue, float)
        return hex(struct.unpack('<Q', struct.pack('<d', fvalue))[0]).rstrip("L")

    @staticmethod
    def hex_to_float(hstr):
        """Returns a float32 value from its hexadecimal representation."""
        assert isinstance(hstr, str)
        return struct.unpack('<f', struct.pack('<I', int(hstr, 16)))[0]

    @staticmethod
    def hex_to_double(hstr):
        """Returns a float64 value from its hexadecimal representation."""

        assert isinstance(hstr, str)
        return struct.unpack('<d', struct.pack('<Q', int(hstr, 16)))[0]

    @staticmethod
    def float_hex_to_double_hex(hstr):
        """Returns the float64 hexadecimal representation from a float32
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        double_value = TestTuple.hex_to_float(hstr)
        return TestTuple.double_to_hex(double_value)

    @staticmethod
    def float_hex_to_inv_double_hex(hstr):
        """Returns the inverted float64 hexadecimal representation from a
           float32 hexadecimal representation.
        """
        assert isinstance(hstr, str)
        temp = TestTuple.hex_to_float(hstr)
        double_value = np.divide(1.0, temp)
        return TestTuple.double_to_hex(double_value)

    @staticmethod
    def float_hex_to_int64_str(hstr):
        """Returns the int64 string representation from a float32
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        x = TestTuple.hex_to_float(hstr)
        if x > np.iinfo(np.dtype('int64')).max:
            return str(np.iinfo(np.dtype('int64')).max)
        if x < np.iinfo(np.dtype('int64')).min:
            return str(np.iinfo(np.dtype('int64')).min)
        return str(int(x))

    @staticmethod
    def float_hex_to_uint64_str(hstr):
        """Returns the uint64 string representation from a float32
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        x = TestTuple.hex_to_float(hstr)
        if x > np.iinfo(np.dtype('uint64')).max:
            return str(np.iinfo(np.dtype('uint64')).max)
        if x < np.iinfo(np.dtype('uint64')).min:
            return str(np.iinfo(np.dtype('uint64')).min)
        return str(int(x))

    @staticmethod
    def int_str_to_bool_str(istr):
        """Returns a bool/integer string from an (arbitrary size) integet string."""
        assert isinstance(istr, str)
        return str(int(bool(int(istr))))

    @staticmethod
    def int_str_to_float_hex(istr):
        """Returns a float32 hexadecimal representation from an (arbitrary size) integer string."""
        assert isinstance(istr, str)
        return TestTuple.float_to_hex(np.float32(int(istr)))

    @staticmethod
    def int_str_to_double_hex(istr):
        """Returns a float64 hexadecimal representation from an (arbitrary size) integer string."""
        assert isinstance(istr, str)
        return TestTuple.double_to_hex(float(istr))

    @staticmethod
    def int_str_to_double_str(istr):
        """Returns a float64 string from an (arbitrary size) integer string."""
        assert isinstance(istr, str)
        return str(float(istr))

    @staticmethod
    def int_str_to_int32_str(istr):
        """Returns an int32 string from an (arbitrary size) integer string."""
        x = int(istr) & (2**32 - 1)
        if x >= 2**31:
            x -= 2**32
        return str(x)

    @staticmethod
    def int_str_to_uint32_str(istr):
        """Returns an uint32 string from an (arbitrary size) integer string."""
        x = int(istr) & (2**32 - 1)
        return str(x)

    @staticmethod
    def int_str_to_int64_str(istr):
        """Returns an int64 string from an (arbitrary size) integer string."""
        x = int(istr) & (2**64 - 1)
        if x >= 2**63:
            x -= 2**64
        return str(x)

    @staticmethod
    def int_str_to_uint64_str(istr):
        """Returns an uint64 string from an (arbitrary size) integer string."""
        x = int(istr) & (2**64 - 1)
        return str(x)

    @staticmethod
    def int_str_to_type_str(target_type):
        """Returns a function for converting an int string to a string for the given type."""
        assert target_type in ('d', 'i64', 'u64')
        if target_type == 'd':
            return TestTuple.int_str_to_double_str
        elif target_type == 'i64':
            return TestTuple.int_str_to_int64_str
        elif target_type == 'u64':
            return TestTuple.int_str_to_uint64_str

    @staticmethod
    def double_hex_to_bool_str(hstr):
        """Returns a bool string from a float64 hexadecimal representation."""
        assert isinstance(hstr, str)
        bool_double = TestTuple.hex_to_double(hstr)
        return '1' if bool_double != 0.0 else '0'

    @staticmethod
    def double_hex_to_int_str(hstr):
        """Returns an int32 string from a float64 hexadecimal
           representation.
        """
        assert isinstance(hstr, str)
        int_double = TestTuple.hex_to_double(hstr)
        if int_double > np.iinfo(np.dtype('int32')).max:
            return str(np.iinfo(np.dtype('int32')).max)
        if int_double < np.iinfo(np.dtype('int32')).min:
            return str(np.iinfo(np.dtype('int32')).min)
        return str(int(int_double))

    @staticmethod
    def double_hex_to_uint_str(hstr):
        """Returns an uint32 string from a float64 hexadecimal
           representation.
        """
        assert isinstance(hstr, str)
        uint_double = TestTuple.hex_to_double(hstr)
        if uint_double > np.iinfo(np.dtype('uint32')).max:
            return str(np.iinfo(np.dtype('uint32')).max)
        if uint_double < np.iinfo(np.dtype('uint32')).min:
            return str(np.iinfo(np.dtype('uint32')).min)
        return str(int(uint_double))

    @staticmethod
    def double_hex_to_float_hex(hstr):
        """Returns the float32 hexadecimal representation from a float64
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        float_double = np.float32(TestTuple.hex_to_double(hstr))
        return TestTuple.float_to_hex(float_double)

    @staticmethod
    def double_hex_to_inv_float_hex(hstr):
        """Returns the inverted float32 hexadecimal representation from a
           float64 hexadecimal representation.
        """
        assert isinstance(hstr, str)
        temp = np.divide(1.0, TestTuple.hex_to_double(hstr))
        float_double = np.float32(temp)
        return TestTuple.float_to_hex(float_double)

    @staticmethod
    def double_hex_to_int64_str(hstr):
        """Returns the int64 string representation from a float64
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        x = TestTuple.hex_to_double(hstr)
        if x > np.iinfo(np.dtype('int64')).max:
            return str(np.iinfo(np.dtype('int64')).max)
        if x < np.iinfo(np.dtype('int64')).min:
            return str(np.iinfo(np.dtype('int64')).min)
        return str(int(x))

    @staticmethod
    def double_hex_to_uint64_str(hstr):
        """Returns the uint64 string representation from a float64
           hexadecimal representation.
        """
        assert isinstance(hstr, str)
        x = TestTuple.hex_to_double(hstr)
        if x > np.iinfo(np.dtype('uint64')).max:
            return str(np.iinfo(np.dtype('uint64')).max)
        if x < np.iinfo(np.dtype('uint64')).min:
            return str(np.iinfo(np.dtype('uint64')).min)
        return str(int(x))

    def __init__(self, ver, stage,
                 first_dimension, second_dimension,
                 basic_type, target_type, names_only):
        assert stage in ('vert', 'geom', 'frag')
        assert first_dimension in ('1', '2', '3', '4')
        assert second_dimension in ('1', '2', '3', '4')
        assert isinstance(names_only, bool)

        self._ver = ver
        self._stage = stage
        self._basic_type = basic_type
        self._target_type = target_type
        self._names_only = names_only
        self._target_full_type = ''
        self._conversion_type = ''
        self._uniform_type = ''
        self._amount = int(first_dimension) * int(second_dimension)
        self._filenames = []
        self._extensions = []

        if ver.startswith('GL_'):
            if basic_type == 'd' or target_type == 'd':
                self._extensions.append('GL_ARB_gpu_shader_fp64')
            if basic_type in ('i64', 'u64') or target_type in ('i64', 'u64'):
                self._extensions.append('GL_ARB_gpu_shader_int64')

        if first_dimension != '1':
            dimensional_type = 'mat' + first_dimension
            if first_dimension != second_dimension:
                dimensional_type += 'x' + second_dimension
        elif second_dimension != '1':
            dimensional_type = 'vec' + second_dimension
        else:
            dimensional_type = ''

        if dimensional_type == '':
            if basic_type == 'b':
                self._conversion_type = 'bool'
                self._uniform_type = 'int'
            elif basic_type == 'i':
                self._conversion_type = 'int'
            elif basic_type == 'u':
                self._conversion_type = 'uint'
            elif basic_type == 'f':
                self._conversion_type = 'float'
            elif basic_type == 'd':
                self._conversion_type = 'double'
            elif basic_type == 'i64':
                self._conversion_type = 'int64_t'
            elif basic_type == 'u64':
                self._conversion_type = 'uint64_t'
            if self._uniform_type == '':
                self._uniform_type = self._conversion_type
            if target_type == 'd':
                self._target_full_type = 'double'
            elif target_type == 'i64':
                self._target_full_type = 'int64_t'
            elif target_type == 'u64':
                self._target_full_type = 'uint64_t'
        else:
            self._conversion_type = (basic_type if basic_type != 'f' else '') + dimensional_type
            if basic_type == 'b':
                self._uniform_type = 'i' + dimensional_type
            else:
                self._uniform_type = self._conversion_type
            self._target_full_type = target_type + dimensional_type

    @abc.abstractmethod
    def _gen_to_target(self):
        """Generates the test files for conversions to float64."""

    @abc.abstractmethod
    def _gen_from_target(self):
        """Generates the test files for conversions from float64."""

    @property
    def filenames(self):
        """Returns the test file names this tuple will generate."""
        if self._filenames == []:
            tmp = self._names_only
            self._names_only = True
            self.generate_test_files()
            self._names_only = tmp
        return self._filenames

    def generate_test_files(self):
        """Generate the GLSL parser tests."""
        self._filenames = []

        self._gen_to_target()
        self._gen_from_target()


class RegularTestTuple(TestTuple):
    """Derived class for conversion tests using regular values within the
       edges of the used types.
    """

    @staticmethod
    def all_tests(names_only):
        """Returns all the possible contained conversion test instances."""

        assert isinstance(names_only, bool)
        stages = ['vert', 'geom', 'frag']
        dimensions = ['1', '2', '3', '4']
        basic_types = ['b', 'u', 'i', 'f', 'd', 'i64', 'u64']
        target_types = ['d', 'i64', 'u64']
        glsl_ver = ['GL_ARB_gpu_shader_int64', 'GL_ARB_gpu_shader_fp64', '400']

        if not names_only:
            test_types = ['compiler', 'execution']
            for ver, test_type in itertools.product(glsl_ver, test_types):
                utils.safe_makedirs(get_dir_name(ver, test_type))

        for ver, stage, first_dimension, second_dimension, basic_type, target_type in itertools.product(
                glsl_ver,
                stages,
                dimensions,
                dimensions,
                basic_types,
                target_types):
            has_int64 = basic_type in ('i64', 'u64') or target_type in ('i64', 'u64')
            if (not (first_dimension != '1' and
                     (second_dimension == '1' or basic_type not in ('f', 'd') or target_type != 'd')) and
                (basic_type not in target_types or basic_type < target_type) and
                ((ver == 'GL_ARB_gpu_shader_int64') == has_int64)):
                yield RegularTestTuple(ver, stage,
                                       first_dimension, second_dimension,
                                       basic_type, target_type, names_only)

    def __init__(self, ver, stage,
                 first_dimension, second_dimension,
                 basic_type, target_type, names_only):
        assert ver in ('GL_ARB_gpu_shader_int64', 'GL_ARB_gpu_shader_fp64', '400')
        assert basic_type in ('b', 'u', 'i', 'f', 'd', 'i64', 'u64')
        assert target_type in ('d', 'i64', 'u64')
        assert not (first_dimension != '1' and
                    (second_dimension == '1' or basic_type not in ('f', 'd') or target_type != 'd'))
        super(RegularTestTuple, self).__init__(ver, stage,
                                               first_dimension, second_dimension,
                                               basic_type, target_type, names_only)

    def _gen_comp_test(self, from_type, to_type, converted_from):
        filename = os.path.join(
            get_dir_name(self._ver, 'compiler'),
            '{}-conversion-implicit-{}-{}-bad.{}'.format(self._stage, from_type, to_type,
                                                         self._stage))

        self._filenames.append(filename)

        if not self._names_only:
            with open(filename, 'w') as test_file:
                test_file.write(TEMPLATES.get_template(
                    'compiler.{}.mako'.format(self._stage)).render_unicode(
                        ver=self._ver,
                        extensions=self._extensions,
                        from_type=from_type,
                        to_type=to_type,
                        converted_from=converted_from))

    def _gen_exec_test(self, from_type, to_type,
                       uniform_from_type, uniform_to_type,
                       explicit, converted_from, conversions):
        filename = os.path.join(
            get_dir_name(self._ver, 'execution'),
            '{}-conversion-{}-{}-{}.shader_test'.format(self._stage, explicit,
                                                        from_type, to_type))

        self._filenames.append(filename)

        if not self._names_only:
            with open(filename, 'w') as test_file:
                test_file.write(TEMPLATES.get_template(
                    'execution.{}.shader_test.mako'.format(self._stage)).render_unicode(
                        ver=self._ver,
                        extensions=self._extensions,
                        amount=self._amount,
                        from_type=from_type,
                        to_type=to_type,
                        converted_from=converted_from,
                        uniform_from_type=uniform_from_type,
                        uniform_to_type=uniform_to_type,
                        conversions=conversions))

    def _gen_to_target(self):
        converted_from = 'from'
        explicit = 'implicit'

        if self._basic_type == 'b':
            conversion_values = BOOL_VALUES
            conversion_function = TestTuple.int_str_to_type_str(self._target_type)
        elif self._basic_type == 'i':
            conversion_values = INT_VALUES
            conversion_function = TestTuple.int_str_to_type_str(self._target_type)
        elif self._basic_type == 'u':
            conversion_values = UINT_VALUES
            conversion_function = TestTuple.int_str_to_type_str(self._target_type)
        elif self._basic_type == 'f':
            if self._target_type == 'd':
                conversion_values = FLOAT_INFS + FLOAT_NEG_ZERO + FLOAT_POS_ZERO + FLOAT_VALUES
                conversion_function = TestTuple.float_hex_to_double_hex
            elif self._target_type == 'i64':
                conversion_values = FLOAT_POS_ZERO + FLOAT_INT64_VALUES
                conversion_function = TestTuple.float_hex_to_int64_str
            elif self._target_type == 'u64':
                conversion_values = FLOAT_POS_ZERO + FLOAT_UINT64_VALUES
                conversion_function = TestTuple.float_hex_to_uint64_str
        elif self._basic_type == 'd':
            if self._target_type == 'i64':
                conversion_values = DOUBLE_DENORMAL_VALUES + DOUBLE_NORMAL_VALUES + DOUBLE_INT_VALUES
                conversion_function = TestTuple.double_hex_to_int64_str
            elif self._target_type == 'u64':
                conversion_values = DOUBLE_DENORMAL_VALUES + DOUBLE_NORMAL_VALUES + DOUBLE_UINT_VALUES
                conversion_function = TestTuple.double_hex_to_uint64_str
        elif self._basic_type in ('i64', 'u64'):
            if self._basic_type == 'i64':
                conversion_values = INT64_VALUES
            else:
                conversion_values = UINT64_VALUES
            conversion_function = TestTuple.int_str_to_type_str(self._target_type)

        if (self._basic_type == 'b' or
            (self._basic_type in ('f', 'd') and self._target_type in ('i64', 'u64')) or
            (self._basic_type == 'u' and self._target_type == 'i64')):
            explicit = 'explicit'
            self._gen_comp_test(self._conversion_type, self._target_full_type,
                                converted_from)
            converted_from = self._target_full_type + '(from)'

        conversions = []
        for value in conversion_values:
            to_value = conversion_function(value)
            item = {'from': value, 'to': to_value}
            conversions.append(item)

        self._gen_exec_test(self._conversion_type, self._target_full_type,
                            self._uniform_type, self._target_full_type,
                            explicit, converted_from, conversions)

    def _gen_from_target(self):
        converted_from = 'from'
        explicit = 'implicit'

        if self._target_type == 'd':
            if self._basic_type == 'b':
                conversion_values = DOUBLE_INFS + DOUBLE_NORMAL_VALUES + DOUBLE_BOOL_VALUES
                conversion_function = TestTuple.double_hex_to_bool_str
            elif self._basic_type == 'i':
                conversion_values = DOUBLE_DENORMAL_VALUES + DOUBLE_NORMAL_VALUES + DOUBLE_INT_VALUES
                conversion_function = TestTuple.double_hex_to_int_str
            elif self._basic_type == 'u':
                conversion_values = DOUBLE_DENORMAL_VALUES + DOUBLE_NORMAL_VALUES + DOUBLE_UINT_VALUES
                conversion_function = TestTuple.double_hex_to_uint_str
            elif self._basic_type == 'f':
                conversion_values = DOUBLE_INFS + DOUBLE_FLOAT_INFS + DOUBLE_FLOAT_VALUES
                conversion_function = TestTuple.double_hex_to_float_hex
            conversion_values = DOUBLE_NEG_ZERO + DOUBLE_POS_ZERO + conversion_values
        elif self._target_type in ('i64', 'u64'):
            if self._target_type == 'i64':
                conversion_values = INT64_VALUES
            elif self._target_type == 'u64':
                conversion_values = UINT64_VALUES

            if self._basic_type == 'b':
                conversion_function = TestTuple.int_str_to_bool_str
            elif self._basic_type == 'i':
                conversion_function = TestTuple.int_str_to_int32_str
            elif self._basic_type == 'u':
                conversion_function = TestTuple.int_str_to_uint32_str
            elif self._basic_type == 'f':
                conversion_function = TestTuple.int_str_to_float_hex
            elif self._basic_type == 'd':
                conversion_function = TestTuple.int_str_to_double_hex
            elif self._basic_type == 'i64':
                conversion_function = TestTuple.int_str_to_int64_str
            elif self._basic_type == 'i':
                conversion_function = TestTuple.int_str_to_uint64_str

        if self._basic_type != 'd':
            self._gen_comp_test(self._target_full_type, self._conversion_type,
                                converted_from)

            converted_from = self._conversion_type + '(from)'
            explicit = 'explicit'
        else:
            assert self._target_type in ('i64', 'u64')

        conversions = []
        for value in conversion_values:
            to_value = conversion_function(value)
            item = {'from': value, 'to': to_value}
            conversions.append(item)

        self._gen_exec_test(self._target_full_type, self._conversion_type,
                            self._target_full_type, self._uniform_type,
                            explicit, converted_from, conversions)


class ZeroSignTestTuple(TestTuple):
    """Derived class for conversion tests using the float32 and float64
       +/-0.0 values.
    """

    @staticmethod
    def all_tests(names_only):
        """Returns all the possible zero sign conversion test instances."""

        assert isinstance(names_only, bool)
        stages = ['vert', 'geom', 'frag']
        dimensions = ['1', '2', '3', '4']
        basic_types = ['f']
        glsl_ver = ['410', '420']

        if not names_only:
            for ver in glsl_ver:
                utils.safe_makedirs(get_dir_name(ver, 'execution'))

        for ver, stage, first_dimension, second_dimension, basic_type in itertools.product(
                glsl_ver,
                stages,
                dimensions,
                dimensions,
                basic_types):
            if not (first_dimension != '1' and second_dimension == '1'):
                yield ZeroSignTestTuple(ver, stage,
                                        first_dimension, second_dimension,
                                        basic_type, names_only)

    def __init__(self, ver, stage,
                 first_dimension, second_dimension,
                 basic_type, names_only):
        assert ver in ('410', '420')
        assert basic_type == 'f'
        assert not (first_dimension != '1' and second_dimension == '1')
        super(ZeroSignTestTuple, self).__init__(ver, stage,
                                                first_dimension, second_dimension,
                                                basic_type, 'd', names_only)

    def __gen_zero_sign_exec_test(self, from_type, to_type,
                                  uniform_from_type, uniform_to_type,
                                  explicit, converted_from, conversions):
        filename = os.path.join(
            get_dir_name(self._ver, 'execution'),
            '{}-conversion-{}-{}-{}-zero-sign.shader_test'.format(self._stage, explicit,
                                                                  from_type, to_type))

        self._filenames.append(filename)

        if not self._names_only:
            with open(filename, 'w') as test_file:
                test_file.write(TEMPLATES.get_template(
                    'execution-zero-sign.{}.shader_test.mako'.format(
                        self._stage)).render_unicode(
                            ver=self._ver,
                            extensions=self._extensions,
                            amount=self._amount,
                            from_type=from_type,
                            to_type=to_type,
                            converted_from=converted_from,
                            uniform_from_type=uniform_from_type,
                            uniform_to_type=uniform_to_type,
                            conversions=conversions))

    def _gen_to_target(self):
        if self._ver == '410':
            conversion_values = FLOAT_POS_ZERO
        elif self._ver == '420':
            conversion_values = FLOAT_NEG_ZERO

        conversions = []
        for value in conversion_values:
            to_value = TestTuple.float_hex_to_inv_double_hex(value)
            item = {'from': value, 'to': to_value}
            conversions.append(item)

        self.__gen_zero_sign_exec_test(self._conversion_type, self._target_full_type,
                                       self._uniform_type, self._target_full_type,
                                       'implicit', 'from', conversions)

    def _gen_from_target(self):
        if self._ver == '410':
            conversion_values = DOUBLE_POS_ZERO
        elif self._ver == '420':
            conversion_values = DOUBLE_NEG_ZERO

        conversions = []
        for value in conversion_values:
            to_value = TestTuple.double_hex_to_inv_float_hex(value)
            item = {'from': value, 'to': to_value}
            conversions.append(item)

        self.__gen_zero_sign_exec_test(self._target_full_type, self._conversion_type,
                                       self._target_full_type, self._uniform_type,
                                       'explicit', self._conversion_type + '(from)', conversions)


def main():
    """Main function."""

    parser = argparse.ArgumentParser(
        description="Generate shader tests that check the conversions from and "
        "to fp64")
    parser.add_argument(
        '--names-only',
        dest='names_only',
        action='store_true',
        default=False,
        help="Don't output files, just generate a list of filenames to stdout")
    args = parser.parse_args()

    np.seterr(divide='ignore')

    for test in (list(RegularTestTuple.all_tests(args.names_only)) +
                 list(ZeroSignTestTuple.all_tests(args.names_only))):
        test.generate_test_files()
        for filename in test.filenames:
            print(filename)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2010, 2014 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Generate tests for builtin const equality tests """

from __future__ import print_function, division, absolute_import
import re
import os

from templates import template_file
from modules import utils

TEMPLATE = template_file(os.path.basename(os.path.splitext(__file__)[0]),
                         'template.shader_test.mako')

TEST_VECTORS = [
    ["vec2(3.0, 3.14)",
     "vec2(-6.0, 7.88)",
     "bvec2(false, false)"
    ],
    ["vec3(13.4, -0.9, 12.55)",
     "vec3(13.4, 12.0, -55.3)",
     "bvec3(true, false, false)"
    ],
    ["vec4(-2.0, 0.0, 0.123, -1000.5)",
     "vec4(-2.4, 0.0, 0.456, 12.5)",
     "bvec4(false, true, false, false)"
    ],
    ["ivec2(-8, 12)",
     "ivec2(-19, 12)",
     "bvec2(false, true)"
    ],
    ["ivec3(0, 8, 89)",
     "ivec3(4, -7, 33)",
     "bvec3(false, false, false)"
    ],
    ["ivec4(11, 1000, 1, -18)",
     "ivec4(55, 1000, -21, -17)",
     "bvec4(false, true, false, false)"
    ],
    ["bvec2(true, false)",
     "bvec2(true, true)",
     "bvec2(true, false)"
    ],
    ["bvec3(false, true, false)",
     "bvec3(false, false, true)",
     "bvec3(true, false, false)"
    ],
    ["bvec4(true, false, false, true)",
     "bvec4(true, true, false, false)",
     "bvec4(true, false, true, false)"
    ],
]


def main():
    """ Main function """
    dirname = os.path.join('spec', 'glsl-1.20', 'execution',
                           'built-in-functions')
    utils.safe_makedirs(dirname)

    for test_id, x in enumerate(TEST_VECTORS, start=2):
        # make equal tests
        name = os.path.join(
            dirname,
            "glsl-const-builtin-equal-{0:02d}.shader_test".format(test_id))

        print(name)

        with open(name, 'w') as f:
            f.write(TEMPLATE.render_unicode(
                func='equal', input=x[0:2], expected=x[2]))

        # make notEqual tests
        name = os.path.join(
            dirname,
            "glsl-const-builtin-notEqual-{0:02d}.shader_test".format(test_id))

        # When generating the notEqual tests, each of the values in the
        # expected result vector need to be inverted
        expected = re.sub("true", "FALSE", x[2])
        expected = re.sub("false", "TRUE", expected)
        expected = expected.lower()

        print(name)

        with open(name, 'w') as f:
            f.write(TEMPLATE.render_unicode(
                func='notEqual', input=x[0:2], expected=expected))


if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (c) 2015 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generate tests for glsl 1.10 and 1.20 variable index reads."""

from __future__ import print_function, absolute_import, division
import os
import itertools

from six.moves import range

from templates import template_dir
from modules.utils import lazy_property, safe_makedirs

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))
FS_TEMPLATE = TEMPLATES.get_template('fs.shader_test.mako')
VS_TEMPLATE = TEMPLATES.get_template('vs.shader_test.mako')
DIRNAME = os.path.join('spec', 'glsl-{}', 'execution', 'variable-indexing')


class TestParams(object):
    """Parameters for a single test.

    This is all of the non-formatting logic of the test. Each property is
    wrapped with a lazy_property decorator, which means the data is cached
    after it is calculated once.

    """
    def __init__(self, matrix_dim, array_dim, mode, index_value, col,
                 expect_type, glsl_version):
        self.matrix_dim = matrix_dim
        self.array_dim = array_dim
        self.mode = mode
        self.index_value = index_value
        self.__col = col
        self.expect_type = expect_type

        assert glsl_version in [110, 120]
        self.glsl_version = glsl_version

    @lazy_property
    def idx(self):
        if self.array_dim != 0:
            return '[{}]'.format(self.index_value)
        else:
            return ''

    @lazy_property
    def cxr_type(self):
        return 'mat{0}x{0}'.format(self.matrix_dim)

    @lazy_property
    def base_type(self):
        if self.glsl_version == 120:
            return self.cxr_type
        else:
            return 'mat{}'.format(self.matrix_dim)

    @lazy_property
    def type(self):
        if self.array_dim != 0 and self.glsl_version == 120:
            return '{}[{}]'.format(self.base_type, self.array_dim)
        else:
            return self.base_type

    @lazy_property
    def dim(self):
        if self.array_dim != 0 and self.glsl_version == 110:
            return '[{}]'.format(self.array_dim)
        else:
            # XXX: should this be an error?
            return ''

    @lazy_property
    def row(self):
        if self.expect_type == 'float':
            return '[row]'
        else:
            # XXX: Should this be an error?
            return ''

    @lazy_property
    def col(self):
        return '[{}]'.format(self.__col)

    @lazy_property
    def test_sizes(self):
        if self.array_dim == 0:
            return [1]
        elif self.index_value == 'index':
            return list(range(1, 1 + self.array_dim))
        else:
            return [2]

    @lazy_property
    def test_columns(self):
        if self.col == '[col]':
            return list(range(1, 1 + self.matrix_dim))
        else:
            return [2]

    @lazy_property
    def test_rows(self):
        if self.expect_type == 'float':
            return list(range(1, 1 + self.matrix_dim))
        else:
            return [1]

    @lazy_property
    def test_array_dim(self):
        if (self.mode == 'uniform' and
                self.glsl_version == 110 and
                self.array_dim != 0 and
                self.index_value != 'index'):
            return self.index_value + 1
        else:
            return self.array_dim

    @lazy_property
    def varying_comps(self):
        if self.array_dim != 0:
            return 4 + self.matrix_dim**2 * self.array_dim
        else:
            return 4 + self.matrix_dim**2

    @lazy_property
    def formated_version(self):
        return '{:.2f}'.format(float(self.glsl_version) / 100)


def make_fs(name, params):
    """Generate a fragment shader test."""
    dirname = DIRNAME.format(params.formated_version)
    safe_makedirs(dirname)
    with open(os.path.join(dirname, name), 'w') as f:
        f.write(FS_TEMPLATE.render_unicode(params=params))
    print(name)


def make_vs(name, params):
    """Generate a vertex shader test."""
    dirname = DIRNAME.format(params.formated_version)
    safe_makedirs(dirname)
    with open(os.path.join(dirname, name), 'w') as f:
        f.write(VS_TEMPLATE.render_unicode(params=params))
    print(name)


def main():
    """Generate the tests."""
    modes = ['temp', 'uniform', 'varying']
    array_dims = [0, 3]
    matrix_dims = [2, 3, 4]
    glsl_versions = [110, 120]
    cols = [1, 'col']
    stages = ['fs', 'vs']
    iter_ = itertools.product(stages, modes, array_dims, matrix_dims,
                              glsl_versions, cols)
    for stage, mode, array_dim, matrix_dim, glsl_version, col in iter_:
        if stage == 'vs':
            func = make_vs
        else:
            func = make_fs

        for expect in ['float', 'vec{}'.format(matrix_dim)]:
            if array_dim != 0:
                arr = 'array-'
                idx_text = 'index-'

                func(
                    '{stage}-{mode}-{arr}mat{matrix_dim}-{col}{row}rd.shader_test'.format(
                        stage=stage,
                        mode=mode,
                        arr=arr,
                        matrix_dim=matrix_dim,
                        col='col-' if col == 'col' else '',
                        row='row-' if expect == 'float' else ''),
                    TestParams(matrix_dim, array_dim, mode, 1, col, expect,
                               glsl_version))
            else:
                arr = ''
                idx_text = ''

            func(
                '{stage}-{mode}-{arr}mat{matrix_dim}-{idx_text}{col}{row}rd.shader_test'.format(
                    stage=stage,
                    mode=mode,
                    arr=arr,
                    matrix_dim=matrix_dim,
                    idx_text=idx_text,
                    col='col-' if col == 'col' else '',
                    row='row-' if expect == 'float' else ''),
                TestParams(matrix_dim, array_dim, mode, 'index', col, expect,
                           glsl_version))


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Parse gl.xml into Python objects.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import os.path
import re
import sys
import functools
from copy import copy, deepcopy

import six

# Export 'debug' so other Piglit modules can easily enable it.
debug = False


def _log_debug(msg):
    if debug:
        print('debug: {0}: {1}'.format(__name__, msg), file=sys.stderr)


# Prefer the external module 'lxml.etree' (it uses libxml2) over Python's
# builtin 'xml.etree.ElementTree'.  It's faster.
try:
    import lxml.etree as etree
    _log_debug('etree is lxml.etree')
except ImportError:
    import xml.etree.cElementTree as etree
    _log_debug('etree is xml.etree.cElementTree')


def parse():
    """Parse gl.xml and return a Registry object."""
    filename = os.path.join(os.path.dirname(__file__), 'gl.xml')
    xml_registry = etree.parse(filename).getroot()
    _repair_xml(xml_registry)
    return Registry(xml_registry)


def _repair_xml(xml_registry):
    fixes = set((
        'GL_ALL_ATTRIB_BITS',
        'gles2_GL_ACTIVE_PROGRAM_EXT',
    ))

    remove_queue = []

    def defer_removal(parent, child):
        remove_queue.append((parent, child))

    for enums in xml_registry.iterfind('./enums'):
        if ('GL_ALL_ATTRIB_BITS' in fixes and
                enums.get('group') == 'AttribMask'):
            # The XML defines GL_ALL_ATTRIB_BITS incorrectly with all bits
            # set (0xFFFFFFFF). From the GL_ARB_multisample spec, v5:
            #
            #     In order to avoid incompatibility with GL implementations
            #     that do not support SGIS_multisample, ALL_ATTRIB_BITS
            #     does not include MULTISAMPLE_BIT_ARB.
            #
            enum = enums.find("./enum[@name='GL_ALL_ATTRIB_BITS']")
            enum.set('value', '0x000FFFFF')

            fixes.remove('GL_ALL_ATTRIB_BITS')
            continue

        if ('gles2_GL_ACTIVE_PROGRAM_EXT' in fixes and
                enums.get('vendor') is not None and
                enums.get('vendor') == 'ARB' and
                enums.get('start') is not None and
                enums.get('start') <= '0x8259' and
                enums.get('end') is not None and
                enums.get('end') >= '0x8259'):
            # GL_ACTIVE_PROGRAM_EXT has different numerical values in GL
            # (0x8B8D) and in GLES (0x8259). Remove the GLES value to avoid
            # redefinition collisions.
            bad_enum = enums.find(("./enum"
                                   "[@value='0x8259']"
                                   "[@name='GL_ACTIVE_PROGRAM_EXT']"
                                   "[@api='gles2']"))
            defer_removal(enums, bad_enum)

            fixes.remove('gles2_GL_ACTIVE_PROGRAM_EXT')
            continue

    for (parent, child) in remove_queue:
        parent.remove(child)

    if len(fixes) > 0:
        raise Exception('failed to apply some xml repairs: ' +
                        ', '.join(map(repr, fixes)))


class OrderedKeyedSet(object):
    """A set with keyed elements that preserves order of element insertion.

    Why waste words? Let's document the class with example code.

    Example:
        Cheese = namedtuple('Cheese', ('name', 'flavor'))
        cheeses = OrderedKeyedSet(key='name')
        cheeses.add(Cheese(name='cheddar', flavor='good'))
        cheeses.add(Cheese(name='gouda', flavor='smells like feet'))
        cheeses.add(Cheese(name='romano', flavor='awesome'))

        # Elements are retrievable by key.
        assert cheeses['gouda'].flavor == 'smells like feet'

        # On key collision, the old element is removed.
        cheeses.add(Cheese(name='gouda', flavor='ok i guess'))
        assert cheeses['gouda'].flavor == 'ok i guess'

        # The set preserves order of insertion. Replacement does not alter
        # order.
        assert list(cheeses)[2].name == 'romano'

        # The set is iterable.
        for cheese in cheeses:
            print(cheese.name)

        # Yet another example...
        Bread = namedtuple('Bread', ('name', 'smell'))
        breads = OrderedKeyedSet(key='name')
        breads.add(Bread(name='como', smell='subtle')
        breads.add(Bread(name='sourdough', smell='pleasant'))

        # The set supports some common set operations, such as union.
        breads_and_cheeses = breads | cheeses
        assert len(breads_and_cheeses) == len(breads) + len(cheeses)
    """

    def __init__(self, key, elems=()):
        """Create a new set with the given key.

        The given 'key' defines how to calculate each element's key value.  If
        'key' is a string, then each key value is defined to be
        `getattr(elem, key)`.  If 'key' is a function, then the key value is
        `key(elem)`.
        """

        # A linked list contains the set's items. Each list node is a 4-tuple
        # [prev, next, key, value].  The root node is permanent.
        root = []
        root[:] = [root, root, None, None]
        self.__list_root = root

        # For quick retrieval, we map each key to its node. That is, each map
        # pair has form {key: [prev, next, key, value])}.
        self.__map = dict()

        if isinstance(key, six.text_type):
            self.__key_func = lambda elem: getattr(elem, key)
        else:
            self.__key_func = key

        for e in elems:
            self.add(e)

    def __or__(self, other):
        """Same as `union`."""
        return self.union(other)

    def __contains__(self, key):
        return key in self.__map

    def __copy__(self):
        return OrderedKeyedSet(key=deepcopy(self.__key_func),
                               elems=iter(self))

    def __getitem__(self, key):
        return self.__map[key][3]

    def __iter__(self):
        return self.itervalues()

    def __len__(self):
        return len(self.__map)

    def __repr__(self):
        templ = '{self.__class__.__name__}({self.name!r})'
        return templ.format(self=self)

    def add(self, value):
        key = self.__key_func(value)
        node = self.__map.get(key, None)
        if node is not None:
            node[3] = value
        else:
            root = self.__list_root
            old_tail = root[0]
            new_tail = [old_tail, root, key, value]
            new_tail[0][1] = new_tail
            new_tail[1][0] = new_tail
            self.__map[key] = new_tail

    def clear(self):
        self.__map.clear()
        root = self.__list_root
        root[:] = [root, root, None, None]

    def extend(self, elems):
        for e in elems:
            self.add(e)

    def get(self, key, default):
        node = self.__map.get(key, None)
        if node is not None:
            return node[3]
        else:
            return default

    def iteritems(self):
        root = self.__list_root
        node = root[1]
        while node is not root:
            yield (node[2], node[3])
            node = node[1]

    def iterkeys(self):
        return (i[0] for i in self.iteritems())

    def itervalues(self):
        return (i[1] for i in self.iteritems())

    def pop(self, key):
        node = self.__map.pop(key)
        node[0][1] = node[1]
        node[1][0] = node[0]
        return node[3]

    def sort_by_key(self):
        sorted_items = sorted(six.iteritems(self.__map))
        self.clear()
        for item in sorted_items:
            self.add(item[1])

    def sort_by_value(self):
        sorted_values = sorted(six.itervalues(self.__map))
        self.clear()
        for value in sorted_values:
            self.add(value)

    def union(self, other):
        """Return the union of two sets as a new set.

        In the new set, all elements of the self set precede those of the other
        set. The order of elements in the new set preserves the order of the
        original sets.

        The new set's key function is copied from self.  On key collisions, set
        y has precedence over x.
        """
        u = copy(self)
        u.extend(other)
        return u


class ImmutableOrderedKeyedSet(OrderedKeyedSet):

    def __init__(self, key, elems):
        self.__is_frozen = False
        OrderedKeyedSet.__init__(self, key=key, elems=elems)
        self.__is_frozen = True

    def add(self, value):
        if self.__is_frozen:
            raise ImmutableError
        else:
            OrderedKeyedSet.add(self, value)

    def pop(self, key):
        raise ImmutableError

    def clear(self):
        raise ImmutableError


class ImmutableError(Exception):
    pass


# Values that may appear in the XML attributes 'api' and 'supported'.
VALID_APIS = frozenset(('gl', 'glcore', 'gles1', 'gles2', 'glsc2'))


class Registry(object):
    """The toplevel <registry> element.

    Attributes:
        features: An OrderedKeyedSet that contains a Feature for each <feature>
        subelement.

        extensions: An OrderedKeyedSet that contains an Extension for each
        <extension> subelement.

        commands: An OrderedKeyedSet that contains a Command for each <command>
        subelement.

        command_alias_map: A CommandAliasMap that contains a CommandAliasSet
        for each equivalence class of commands.

        enum_groups: An OrderedKeyedSet that contains an EnumGroup for each
        <enums> subelement.

        enums: An OrderedKeyedSet that contains an Enum for each <enum>
        subelement.

        vendor_namespaces: A collection of all vendor prefixes and suffixes,
        such as "ARB", "EXT", "CHROMIUM", and "NV".
    """

    def __init__(self, xml_registry):
        """Parse the <registry> element."""

        assert xml_registry.tag == 'registry'

        self.command_alias_map = CommandAliasMap()
        self.commands = OrderedKeyedSet(key='name')
        self.enum_groups = []
        self.enums = OrderedKeyedSet(key='name')
        self.extensions = OrderedKeyedSet(key='name')
        self.features = OrderedKeyedSet(key='name')
        self.vendor_namespaces = set()

        for xml_command in xml_registry.iterfind('./commands/command'):
            command = Command(xml_command)
            self.commands.add(command)
            self.command_alias_map.add(command)

        for xml_enums in xml_registry.iterfind('./enums'):
            enum_group = EnumGroup(xml_enums)
            self.enum_groups.append(enum_group)
            for enum in enum_group.enums:
                self.enums.add(enum)

        for xml_feature in xml_registry.iterfind('./feature'):
            feature = Feature(xml_feature, command_map=self.commands,
                              enum_map=self.enums)
            self.features.add(feature)

        for xml_ext in xml_registry.iterfind('./extensions/extension'):
            ext = Extension(xml_ext, command_map=self.commands,
                            enum_map=self.enums)
            self.extensions.add(ext)
            self.vendor_namespaces.add(ext.vendor_namespace)

        self.vendor_namespaces.remove(None)


@functools.total_ordering
class Feature(object):
    """A <feature> XML element.

    Attributes:
        name: The XML element's 'name' attribute.

        api: The XML element's 'api' attribute.

        version_str:  The XML element's 'number' attribute. For example, "3.1".

        version_float: float(version_str)

        version_int: int(10 * version_float)

        requirements: A collection of Requirement for each Command and Enum
        this Feature requires.
    """

    def __init__(self, xml_feature, command_map, enum_map):
        """Parse a <feature> element."""

        # Example <feature> element:
        #
        #    <feature api="gles2" name="GL_ES_VERSION_3_1" number="3.1">
        #        <!-- arrays_of_arrays features -->
        #        <require/>
        #        <!-- compute_shader features -->
        #        <require>
        #            <command name="glDispatchCompute"/>
        #            <command name="glDispatchComputeIndirect"/>
        #            <enum name="GL_COMPUTE_SHADER"/>
        #            <enum name="GL_MAX_COMPUTE_UNIFORM_BLOCKS"/>
        #            ...
        #        </require>
        #        <!-- draw_indirect features -->
        #        <require>
        #            <command name="glDrawArraysIndirect"/>
        #            <command name="glDrawElementsIndirect"/>
        #            <enum name="GL_DRAW_INDIRECT_BUFFER"/>
        #            <enum name="GL_DRAW_INDIRECT_BUFFER_BINDING"/>
        #        </require>
        #        ...
        #    </feature>

        assert xml_feature.tag == 'feature'

        # Parse the <feature> tag's attributes.
        self.name = xml_feature.get('name')
        self.api = xml_feature.get('api')
        self.is_gles = self.name.startswith('GL_ES')

        self.version_str = xml_feature.get('number')
        self.version_float = float(self.version_str)
        self.version_int = int(10 * self.version_float)

        self.__parse_requirements(xml_feature, command_map, enum_map)

        assert self.api in VALID_APIS
        assert len(self.requirements) > 0

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(other, Extension):
            return False
        elif self.is_gles != other.is_gles:
            return False
        return self.name == other.name

    def __lt__(self, other):
        if isinstance(other, Extension):
            return True
        # Desktop GL before GLES
        elif self.is_gles != other.is_gles:
            if self.is_gles:
                return False
            else:
                return True
        return self.name < other.name

    def __repr__(self):
        templ = '{self.__class__.__name__}({self.name!r})'
        return templ.format(self=self)

    def __parse_requirements(self, xml_feature, command_map, enum_map):
        """For each <command> and <enum> under a <require>, create
        a Requirement that links this Feature to a Command or Enum.
        """
        self.requirements = set()

        def link(x):
            req = Requirement(provider=self, provided=x,
                              apis=frozenset((self.api,)))
            self.requirements.add(req)
            x.requirements.add(req)

        for xml_cmd in xml_feature.iterfind('./require/command'):
            cmd = command_map[xml_cmd.get('name')]
            link(cmd)
        for xml_enum in xml_feature.iterfind('./require/enum'):
            enum = enum_map[xml_enum.get('name')]
            link(enum)


@functools.total_ordering
class Extension(object):
    """An <extension> XML element.

    Attributes:
        name: The XML element's 'name' attribute.

        supported_apis: The set of api strings in the XML element's 'supported'
        attribute. For example, set('gl', 'glcore').

        vendor_namespace: For example, "AMD". May be None.

        requirements: A collection of Requirement for each Command and Enum
        this Extension requires.
    """

    __VENDOR_REGEX = re.compile(r'^GL_(?P<vendor_namespace>[A-Z]+)_')
    RATIFIED_NAMESPACES = ('KHR', 'ARB', 'OES')

    def __init__(self, xml_extension, command_map, enum_map):
        """Parse an <extension> element."""

        # Example <extension> element:
        #     <extension name="GL_ARB_ES2_compatibility" supported="gl|glcore">
        #         <require>
        #             <enum name="GL_FIXED"/>
        #             <enum name="GL_IMPLEMENTATION_COLOR_READ_TYPE"/>
        #             ...
        #             <command name="glReleaseShaderCompiler"/>
        #             <command name="glShaderBinary"/>
        #             ...
        #         </require>
        #     </extension>

        assert xml_extension.tag == 'extension'

        self.name = xml_extension.get('name')

        self.vendor_namespace = None
        match = Extension.__VENDOR_REGEX.match(self.name)
        if match is not None:
            groups = match.groupdict()
            self.vendor_namespace = groups.get('vendor_namespace', None)

        self.supported_apis = xml_extension.get('supported').split('|')
        self.supported_apis = frozenset(self.supported_apis)
        assert self.supported_apis <= VALID_APIS

        self.__parse_requirements(xml_extension, command_map, enum_map)

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(other, Feature):
            return False
        elif self.is_ratified != other.is_ratified:
            return False
        elif self.vendor_namespace == 'EXT' != other.vendor_namespace == 'EXT':
            return False
        return self.name == other.name

    def __lt__(self, other):
        if isinstance(other, Feature):
            return False
        elif self.is_ratified != other.is_ratified:
            # sort ratified before unratified
            if self.is_ratified:
                return True
            else:
                return False
        elif (other.vendor_namespace == 'EXT') != \
                (self.vendor_namespace == 'EXT'):
            # Sort EXT before others
            if self.vendor_namespace == 'EXT':
                return True
            else:
                return False
        return self.name < other.name

    def __repr__(self):
        templ = '{self.__class__.__name__}(name={self.name!r})'
        return templ.format(self=self)

    @property
    def is_ratified(self):
        """True if the vendor namespace is one that traditionally requires
        ratification by Khronos.
        """
        return self.vendor_namespace in self.RATIFIED_NAMESPACES

    def __parse_requirements(self, xml_extension, command_map, enum_map):
        """For each <command> and <enum> under a <require>, create
        a Requirement that links this Extension to a Command or Enum.
        """
        self.requirements = set()

        def link(xml_require, x):
            api = xml_require.get('api', None)
            if api is not None:
                assert api in self.supported_apis
                apis = frozenset((api,))
            else:
                apis = frozenset(self.supported_apis)

            req = Requirement(provider=self, provided=x, apis=apis)
            self.requirements.add(req)
            x.requirements.add(req)

        for xml_req in xml_extension.iterfind('./require'):
            for xml_cmd in xml_req.iterfind('./command'):
                cmd = command_map[xml_cmd.get('name')]
                link(xml_req, cmd)
            for xml_enum in xml_req.iterfind('./enum'):
                enum = enum_map[xml_enum.get('name')]
                link(xml_req, enum)


@functools.total_ordering
class Requirement(object):
    """A <require> XML element, which links a provider (Feature or Extension)
    to a provided (Command or Enum) for a set of apis.
    """

    def __init__(self, provider, provided, apis):
        self.provider = provider
        self.provided = provided
        self.apis = frozenset(apis)

        def choose_if(condition, obj):
            if condition:
                return obj
            else:
                return None

        self.has_feature = isinstance(provider, Feature)
        self.has_extension = isinstance(provider, Extension)
        self.has_command = isinstance(provided, Command)
        self.has_enum = isinstance(provided, Enum)

        self.feature = choose_if(self.has_feature, self.provider)
        self.extension = choose_if(self.has_extension, self.provider)
        self.command = choose_if(self.has_command, self.provided)
        self.enum = choose_if(self.has_enum, self.provided)

        assert self.has_feature + self.has_extension == 1
        assert self.has_command + self.has_enum == 1
        assert self.apis <= VALID_APIS

        _log_debug('created {0}'.format(self))

    def __hash__(self):
        return hash('__Requirement_class_{}_{}_'.format(
            self.provider, self.provided))

    def __eq__(self, other):
        if self.provider != other.provider:
            return False
        elif self.provided != other.provided:
            return False
        return True

    def __lt__(self, other):
        if self.provider < other.provider:
            return True
        elif self.provided < other.provided:
            return True
        return False

    def __repr__(self):
        templ = ('{self.__class__.__name__}'
                 '(provider={self.provider.name!r},'
                 ' provided={self.provided.name!r},'
                 ' apis={api_tuple})')
        return templ.format(self=self, api_tuple=tuple(self.apis))


class CommandParam(object):
    """A <param> XML element at path command/param.

    Attributes:
        name
        c_type
        array_suffix
    """

    __PARAM_NAME_FIXES = {'near': 'hither', 'far': 'yon'}

    def __init__(self, xml_param, log=None):
        """Parse a <param> element."""

        # Example <param> elements:
        #
        #    <param>const <ptype>GLchar</ptype> *<name>name</name></param>
        #    <param len="1"><ptype>GLsizei</ptype> *<name>length</name></param>
        #    <param len="bufSize">
        #        <ptype>GLint</ptype> *<name>values</name>
        #    </param>
        #    <param><ptype>GLenum</ptype> <name>shadertype</name></param>
        #    <param group="sync">
        #        <ptype>GLsync</ptype> <name>sync</name>
        #    </param>
        #    <param><ptype>GLuint</ptype> <name>baseAndCount</name>[2]</param>

        assert xml_param.tag == 'param'

        self.name = xml_param.find('./name').text

        # Rename the parameter if its name is a reserved keyword in MSVC.
        self.name = self.__PARAM_NAME_FIXES.get(self.name, self.name)

        # Parse the C type.
        c_type_text = list(xml_param.itertext())

        # Could be <name> or <array_suffix>
        c_type_text_end = c_type_text.pop(-1)

        # We popped off <array_suffix>
        if c_type_text_end.startswith('['):
            # This is an array variable.
            self.array_suffix = c_type_text_end
            c_type_text.pop(-1)  # Pop off the next one (<name>)
        else:
            self.array_suffix = ''
        c_type_text = (t.strip() for t in c_type_text)
        self.c_type = ' '.join(c_type_text).strip()

        _log_debug('parsed {0}'.format(self))

    def __repr__(self):
        templ = ('{self.__class__.__name__}'
                 '(name={self.name!r}, type={self.c_type!r}, '
                 'suffix={self.array_suffix!r})')
        return templ.format(self=self)


@functools.total_ordering
class Command(object):
    """A <command> XML element.

    Attributes:
        name: The XML element's 'name' attribute, which is also the function
        name.

        c_return_type: For example, "void *".

        alias: The XML element's 'alias' element. May be None.

        param_list: List of that contains a CommandParam for each <param>
        subelement.

        requirements: A collection of each Requirement that exposes this
        Command.
    """

    def __init__(self, xml_command):
        """Parse a <command> element."""

        # Example <command> element:
        #
        #    <command>
        #        <proto>void <name>glTexSubImage2D</name></proto>
        #        <param group="TextureTarget">
        #            <ptype>GLenum</ptype> <name>target</name>
        #        </param>
        #        <param group="CheckedInt32">
        #            <ptype>GLint</ptype> <name>level</name>
        #        </param>
        #        <param group="CheckedInt32">
        #            <ptype>GLint</ptype> <name>xoffset</name>
        #        </param>
        #        <param group="CheckedInt32">
        #            <ptype>GLint</ptype> <name>yoffset</name>
        #        </param>
        #        <param><ptype>GLsizei</ptype> <name>width</name></param>
        #        <param><ptype>GLsizei</ptype> <name>height</name></param>
        #        <param group="PixelFormat">
        #            <ptype>GLenum</ptype> <name>format</name>
        #        </param>
        #        <param group="PixelType">
        #            <ptype>GLenum</ptype> <name>type</name>
        #        </param>
        #        <param len="COMPSIZE(format,type,width,height)">const void *
        #            <name>pixels</name>
        #        </param>
        #        <glx type="render" opcode="4100"/>
        #        <glx type="render" opcode="332" name="glTexSubImage2DPBO"
        #             comment="PBO protocol"/>
        #    </command>
        #

        assert xml_command.tag == 'command'
        xml_proto = xml_command.find('./proto')
        self.name = xml_proto.find('./name').text
        _log_debug('start parsing Command(name={0!r})'.format(self.name))

        self.requirements = set()
        self.__vendor_namespace = None

        # Parse the return type from the <proto> element.
        #
        # Example of a difficult <proto> element:
        #     <proto group="String">const <ptype>GLubyte</ptype> *
        #         <name>glGetStringi</name>
        #     </proto>
        c_return_type_text = list(xml_proto.itertext())
        # Pop off the text from the <name> subelement.
        c_return_type_text.pop(-1)
        c_return_type_text = (t.strip() for t in c_return_type_text)
        self.c_return_type = ' '.join(c_return_type_text).strip()

        # Parse alias info, if any.
        xml_alias = xml_command.find('./alias')
        if xml_alias is None:
            self.alias = None
        else:
            self.alias = xml_alias.get('name')

        self.param_list = [
            CommandParam(xml_param)
            for xml_param in xml_command.iterfind('./param')
        ]

        _log_debug(('parsed {self.__class__.__name__}('
                    'name={self.name!r}, '
                    'alias={self.alias!r}, '
                    'prototype={self.c_prototype!r})').format(self=self))

    def __hash__(self):
        return hash('__Command_class_{}_'.format(self.name))

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        templ = '{self.__class__.__name__}({self.name!r})'
        return templ.format(self=self)

    @property
    def vendor_namespace(self):
        if self.__vendor_namespace is None:
            for req in self.requirements:
                ext = req.extension
                if ext is None:
                    continue

                if ext.vendor_namespace is None:
                    continue

                if self.name.endswith('_' + ext.vendor_namespace):
                    self.__vendor_namespace = ext.vendor_namespace

        return self.__vendor_namespace

    @property
    def c_prototype(self):
        """For example, "void glAccum(GLenum o, GLfloat value)"."""
        return '{self.c_return_type} {self.name}({self.c_named_param_list})'\
            .format(self=self)

    @property
    def c_funcptr_typedef(self):
        """For example, "PFNGLACCUMROC" for glAccum."""
        return 'PFN{0}PROC'.format(self.name).upper()

    @property
    def c_named_param_list(self):
        """For example, "GLenum op, GLfloat value" for glAccum."""
        return ', '.join(
            '{p.c_type} {p.name}{p.array_suffix}'.format(p=param)
            for param in self.param_list
        )

    @property
    def c_unnamed_param_list(self):
        """For example, "GLenum, GLfloat" for glAccum."""
        return ', '.join(
            '{param.c_type}{param.array_suffix}'.format(param=param)
            for param in self.param_list
        )

    @property
    def c_untyped_param_list(self):
        """For example, "op, value" for glAccum."""
        return ', '.join(
            param.name
            for param in self.param_list
        )


@functools.total_ordering
class CommandAliasSet(ImmutableOrderedKeyedSet):

    def __init__(self, commands):
        ImmutableOrderedKeyedSet.__init__(self, key='name',
                                          elems=sorted(commands))
        self.__primary_command = None
        self.__requirements = None

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        templ = '{self.__class__.__name__}({self.name!r})'
        return templ.format(self=self)

    def __hash__(self):
        return hash(repr(self))

    @property
    def name(self):
        return self.primary_command.name

    @property
    def primary_command(self):
        """The set's first command when sorted by name."""
        for command in self:
            return command

    @property
    def requirements(self):
        """A sorted iterator over each Requirement that exposes this
        CommandAliasSet.
        """
        if self.__requirements is None:
            self.__requirements = sorted(
                req
                for command in self
                for req in command.requirements
            )
            _log_debug('{0} sorted requirements: {1}'.format(
                self, self.__requirements))

        return iter(self.__requirements)


class CommandAliasMap(object):

    def __init__(self):
        self.__map = dict()
        self.__sorted_unique_values = None

    def __getitem__(self, command_name):
        return self.__map[command_name]

    def __iter__(self):
        """A sorted iterator over the map's unique CommandAliasSet values."""
        if self.__sorted_unique_values is None:
            self.__sorted_unique_values = \
                sorted(set(six.itervalues(self.__map)))

        return iter(self.__sorted_unique_values)

    def get(self, command_name, default):
        return self.__map.get(command_name, default)

    def add(self, command):
        assert isinstance(command, Command)
        name = command.name
        _log_debug('adding command {0!r} to CommandAliasMap'.format(name))

        name_set = self.get(name, None)
        assert self.__is_set_mapping_complete(name_set)

        alias_set = self.get(command.alias, None)
        assert self.__is_set_mapping_complete(alias_set)

        if name_set is alias_set and name_set is not None:
            return

        # After modifying the contained alias sets, the mapping will no longer
        # be sorted.
        self.__sorted_unique_values = None

        new_set_elems = set((command,))
        if name_set is not None:
            new_set_elems.update(name_set)
        if alias_set is not None:
            new_set_elems.update(alias_set)

        new_set = CommandAliasSet(new_set_elems)
        for other_command in new_set:
            self.__map[other_command.name] = new_set
            if other_command.alias is not None:
                self.__map[other_command.alias] = new_set

    def __is_set_mapping_complete(self, alias_set):
        if alias_set is None:
            return True

        for command in alias_set:
            if self[command.name] is not alias_set:
                return False
            if command.alias is None:
                continue
            if self[command.alias] is not alias_set:
                return False

        return True


class EnumGroup(object):
    """An <enums> element at path registry/enums.

    Attributes:
        name: The XML element's 'group' attribute. If the XML does not define
        'group', then this class invents one.

        type: The XML element's 'type' attribute. If the XML does not define
        'type', then this class invents one.

        start, end: The XML element's 'start' and 'end' attributes. Each may be
        None.

        enums: An OrderedKeyedSet of Enum that contains each <enum> subelement
        in this <enums>.
    """

    # Each EnumGroup belongs to exactly one member of EnumGroup.TYPES.
    #
    # Some members in EnumGroup.TYPES are invented and not present in gl.xml.
    # The only enum type defined explicitly in gl.xml is "bitmask", which
    # occurs as <enums type="bitmask">.  However, in gl.xml each block of
    # non-bitmask enums is introduced by a comment that describes the block's
    # "type", even if the <enums> tag lacks a 'type' attribute. (Thanks,
    # Khronos, for encoding data in XML comments rather than the XML itself).
    # EnumGroup.TYPES lists such implicit comment-only types, with invented
    # names, alongside the types explicitly defined by <enums type=>.
    TYPES = (
        # Type 'default_namespace' is self-explanatory. It indicates the large
        # set of enums from 0x0000 to 0xffff that includes, for example,
        # GL_POINTS and GL_TEXTURE_2D.
        'default_namespace',

        # Type 'bitmask' is self-explanatory.
        'bitmask',

        # Type 'small_index' indicates a small namespace of non-bitmask enums.
        # As of Khronos revision 26792, 'small_index' groups generally contain
        # small numbers used for indexed access.
        'small_index',

        # Type 'special' is used only for the group named "SpecialNumbers". The
        # group contains enums such as GL_FALSE, GL_ZERO, and GL_INVALID_INDEX.
        'special',
    )

    def __init__(self, xml_enums):
        """Parse an <enums> element."""

        # Example of a bitmask group:
        #
        #     <enums namespace="GL" group="SyncObjectMask" type="bitmask">
        #         <enum value="0x00000001"
        #               name="GL_SYNC_FLUSH_COMMANDS_BIT"/>
        #         <enum value="0x00000001"
        #               name="GL_SYNC_FLUSH_COMMANDS_BIT_APPLE"/>
        #     </enums>
        #
        # Example of a group that resides in OpenGL's default enum namespace:
        #
        #     <enums namespace="GL" start="0x0000" end="0x7FFF" vendor="ARB"
        #            comment="...">
        #         <enum value="0x0000" name="GL_POINTS"/>
        #         <enum value="0x0001" name="GL_LINES"/>
        #         <enum value="0x0002" name="GL_LINE_LOOP"/>
        #         ...
        #     </enums>
        #
        # Example of a non-bitmask group that resides outside OpenGL's default
        # enum namespace:
        #
        #     <enums namespace="GL" group="PathRenderingTokenNV" vendor="NV">
        #         <enum value="0x00" name="GL_CLOSE_PATH_NV"/>
        #         <enum value="0x02" name="GL_MOVE_TO_NV"/>
        #         <enum value="0x03" name="GL_RELATIVE_MOVE_TO_NV"/>
        #         ...
        #     </enums>
        #

        self.name = xml_enums.get('group', None)
        _log_debug('start parsing {0}'.format(self))

        self.type = xml_enums.get('type', None)
        self.start = xml_enums.get('start', None)
        self.end = xml_enums.get('end', None)
        self.enums = []

        self.__invent_name_and_type()
        assert self.name is not None
        assert self.type in self.TYPES

        _log_debug('start parsing <enum> subelements of {0}'.format(self))
        self.enums = OrderedKeyedSet(key='name')
        for xml_enum in xml_enums.iterfind('./enum'):
            self.enums.add(Enum(self, xml_enum))
        _log_debug('parsed {0}'.format(self))

    def __repr__(self):
        templ = '{self.__class__.__name__}({self.name!r})'
        return templ.format(self=self)

    def __invent_name_and_type(self):
        """If the XML didn't define a name or type, invent one."""
        if self.name is None:
            assert self.type is None
            assert self.start is not None
            assert self.end is not None
            self.name = 'range_{self.start}_{self.end}'.format(self=self)
            self.type = 'default_namespace'
        elif self.type is None:
            self.type = 'small_index'
        elif self.name == 'SpecialNumbers':
            assert self.type is None
            self.type = 'special'


@functools.total_ordering
class Enum(object):
    """An <enum> XML element.

    Attributes:
        name, api: The XML element's 'name' and 'api' attributes.

        str_value, c_num_literal: Equivalent attributes. The XML element's
        'value' attribute.

        num_value: The long integer for str_value.

        requirements: A collection of each Requirement that exposes this Enum.
    """

    def __init__(self, enum_group, xml_enum):
        """Parse an <enum> tag located at path registry/enums/enum."""

        # Example <enum> element:
        #     <enum value="0x0000" name="GL_POINTS"/>

        assert isinstance(enum_group, EnumGroup)
        assert xml_enum.tag == 'enum'

        self.requirements = set()
        self.__vendor_namespace = None

        self.group = enum_group
        self.name = xml_enum.get('name')
        self.api = xml_enum.get('api')
        self.str_value = xml_enum.get('value')
        self.c_num_literal = self.str_value

        if '0x' in self.str_value.lower():
            base = 16
        else:
            base = 10

        if six.PY2:
            # long is undefined in python3, and we are aware of that
            # pylint: disable=undefined-variable
            self.num_value = long(self.str_value, base)
        else:
            assert six.PY3
            self.num_value = int(self.str_value, base)

        _log_debug('parsed {0}'.format(self))

    def __repr__(self):
        templ = ('{self.__class__.__name__}'
                 '(name={self.name!r},'
                 ' value={self.str_value!r})')
        return templ.format(self=self)

    def __eq__(self, other):
        if self.num_value != other.num_value:
            return False
        elif (self.vendor_namespace is None) != \
                (other.vendor_namespace is None):
            return False
        elif (self.vendor_namespace in Extension.RATIFIED_NAMESPACES) != \
                (other.vendor_namespace in Extension.RATIFIED_NAMESPACES):
            return False
        elif (self.vendor_namespace == 'EXT') != \
                (other.vendor_namespace == 'EXT'):
            return False
        elif self.name != other.name:
            return False
        return self.api == other.api

    def __lt__(self, other):  # pylint: disable=too-many-return-statements
        """Less than.

        Sort by numerical value, then vendor_namspace (ratified first, then
        EXT), then by full name, and finally by api.

        This sort order ensures that names provided by core specifications
        precede those provided by ratified extensions, which proceed those
        provided by unratified extensions.

        For example: GL_RED < GL_RED_EXT < GL_RED_INTEL

        """
        if self.num_value != other.num_value:
            if self.num_value < other.num_value:
                return True
            return False

        x = self.vendor_namespace is None
        y = other.vendor_namespace is None
        if x != y:
            if x and not y:
                return True
            return False

        x = self.vendor_namespace in Extension.RATIFIED_NAMESPACES
        y = other.vendor_namespace in Extension.RATIFIED_NAMESPACES
        if x != y:
            if x and not y:
                return True
            return False

        x = self.vendor_namespace == 'EXT'
        y = other.vendor_namespace == 'EXT'
        if x != y:
            if x and not y:
                return True
            return False

        if self.name != other.name:
            if self.name < other.name:
                return True
            return False

        return self.api < other.api

    @property
    def vendor_namespace(self):
        if self.__vendor_namespace is None:
            for req in self.requirements:
                ext = req.extension
                if ext is None:
                    continue

                if ext.vendor_namespace is None:
                    continue

                if self.name.endswith('_' + ext.vendor_namespace):
                    self.__vendor_namespace = ext.vendor_namespace

        return self.__vendor_namespace



# coding=utf-8
# Copyright (c) 2014 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper functions for test generators."""

from __future__ import print_function, absolute_import
import os
import errno
import functools


def safe_makedirs(dirs):
    """A safe function for creating a directory tree.

    This function wraps os.makedirs, and provides a couple of sanity checks,
    first, if the directory already exists it doesn't try to create it, and
    second if the directory comes into existence between the check and creation
    time (like if another generator creates it) then the exception will be
    caught.

    """
    if not os.path.exists(dirs):
        try:
            os.makedirs(dirs)
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise


class lazy_property(object):
    """Decorator for lazy property loading.

    A property decorated with this class will execute it's code the first time
    it's run, but will store the value after that. This is useful for values
    that are either 1) expensive to compute, or 2) need to be computed only
    once and then read multiple times

    """
    def __init__(self, func):
        functools.wraps(func)
        self.__func = func

    def __get__(self, obj, cls):
        if not obj:
            return self
        value = self.__func(obj)
        setattr(obj, self.__func.__name__, value)
        return value



# -*- coding: utf-8 -*-

"""A quicker profile than all.

This profile filters out a number of very slow tests, and tests that are very
exhaustively tested, since they add a good deal of runtime to piglit.

There are 18000+ auto-generated tests that are exhaustive, but for quick.py we
don't want that level of exhaustiveness, so this filter removes 80% in a random
(but deterministic) way.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from framework import grouptools
from framework.test import PiglitGLTest
from tests.opengl import profile as _profile

__all__ = ['profile']

# See the note in all.py about this warning
# pylint: disable=bad-continuation


profile = _profile.copy()  # pylint: disable=invalid-name

# Set the --quick flag on a few image_load_store_tests
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'arb_shader_image_load_store')) as g:
    with profile.test_list.allow_reassignment:
        g(['arb_shader_image_load_store-coherency', '--quick'], 'coherency')
        g(['arb_shader_image_load_store-host-mem-barrier', '--quick'],
          'host-mem-barrier')
        g(['arb_shader_image_load_store-max-size', '--quick'], 'max-size')
        g(['arb_shader_image_load_store-semantics', '--quick'], 'semantics')
        g(['arb_shader_image_load_store-shader-mem-barrier', '--quick'],
          'shader-mem-barrier')

# Set the --quick flag on the image_size test
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'arb_shader_image_size')) as g:
    with profile.test_list.allow_reassignment:
        g(['arb_shader_image_size-builtin', '--quick'], 'builtin')

# Set the --quick flag on the texture env combine test
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'ext_texture_env_combine')) as g:
    with profile.test_list.allow_reassignment:
        g(['ext_texture_env_combine-combine', '--quick'], 'texture-env-combine')

# Set the --quick flag on the gl-1.0 blending test
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', '!opengl 1.0')) as g:
    with profile.test_list.allow_reassignment:
        g(['gl-1.0-blend-func', '--quick'], 'gl-1.0-blend-func')

# Set the --quick flag on the GL_ARB_sample_locations test
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'arb_sample_locations')) as g:
    with profile.test_list.allow_reassignment:
        g(['arb_sample_locations', '--quick'], 'test')

# Limit texture size to 512x512 for some texture_multisample tests.
# The default (max supported size) can be pretty slow.
with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'ARB_texture_multisample')) as g:
    with profile.test_list.allow_reassignment:
        size_arg = ['--texsize', '512']
        g(['arb_texture_multisample-large-float-texture'] + size_arg,
          'large-float-texture', run_concurrent=False)
        g(['arb_texture_multisample-large-float-texture', '--array'] +
          size_arg, 'large-float-texture-array', run_concurrent=False)
        g(['arb_texture_multisample-large-float-texture', '--fp16'] +
          size_arg, 'large-float-texture-fp16', run_concurrent=False)
        g(['arb_texture_multisample-large-float-texture', '--array',
           '--fp16'] + size_arg,
          'large-float-texture-array-fp16', run_concurrent=False)



# coding=utf-8
# Copyright (c) 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A small library that contains libraries equivalent to six

This function is pending upstreaming in six.

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import six

def python_2_bool_compatible(class_):
    """A decorator to fix __bool__/__nonzero__ name changes."""
    if six.PY2:
        try:
            class_.__nonzero__ = class_.__bool__
        except AttributeError:
            raise ValueError(
                "@python_2_bool_compatible cannot be applied to {} because "
                "it doesn't define __bool__().".format(class_.__name__))
    return class_


# Some version of six don't have this function
try:
    from six import python_2_unicode_compatible
except ImportError:
    def python_2_unicode_compatible(class_):
        """A decorator to fix __str/__bytes__/__unicode__ name changes."""
        if six.PY2:
            failed = False
            try:
                class_.__unicode__ = class_.__str__
            except AttributeError:
                failed = True

            try:
                class_.__str__ = class_.__bytes__
            except AttributeError:
                if failed:
                    raise ValueError(
                        "@python_2_unicode_compatible cannot be applied to {} "
                        "because it doesn't define __str__() "
                        "or __bytes__().".format(class_.__name__))
        return class_


try:
    from six import viewvalues
except ImportError:
    if six.PY2:
        viewvalues = lambda d: d.viewvalues()
    elif six.PY3:
        viewvalues = lambda d: d.values()
